"""
    verktyg.http.cookies
    ~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import re
import string
from itertools import chain
from time import time
from datetime import timedelta

from verktyg.urls import iri_to_uri, encode_idna
from verktyg import datastructures
from verktyg.http.basic import _dump_date


_cookie_params = {
    b'expires', b'path', b'comment', b'max-age',
    b'secure', b'httponly', b'version',
}
_legal_cookie_chars = (
    string.ascii_letters + string.digits + u"!#$%&'*+-.^_`|~:").encode('ascii')

_cookie_quoting_map = {
    b','[0]: b'\\054',
    b';'[0]: b'\\073',
    b'"'[0]: b'\\"',
    b'\\'[0]: b'\\\\',
}
for _i in chain(range(32), range(127, 256)):
    _cookie_quoting_map[_i] = (
        '\\%03o' % _i
    ).encode('latin1')

_octal_re = re.compile(b'\\\\[0-3][0-7][0-7]')
_quote_re = re.compile(b'[\\\\].')
_cookie_re = re.compile(b"""(?x)
    (?P<key>[^=]+)
    \s*=\s*
    (?P<val>
        "(?:[^\\\\"]|\\\\.)*" |
         (?:.*?)
    )
    \s*;
""")


def _cookie_quote(b):
    buf = bytearray()
    all_legal = True

    for char in b:
        if char not in _legal_cookie_chars:
            all_legal = False
            buf.extend(_cookie_quoting_map.get(char, [char]))
        else:
            buf.append(char)

    if all_legal:
        return bytes(buf)
    return bytes(b'"' + buf + b'"')


def _cookie_unquote(b):
    if len(b) < 2:
        return b
    if b[:1] != b'"' or b[-1:] != b'"':
        return b

    b = b[1:-1]

    i = 0
    n = len(b)
    rv = bytearray()
    _push = rv.extend

    while 0 <= i < n:
        o_match = _octal_re.search(b, i)
        q_match = _quote_re.search(b, i)
        if not o_match and not q_match:
            rv.extend(b[i:])
            break
        j = k = -1
        if o_match:
            j = o_match.start(0)
        if q_match:
            k = q_match.start(0)
        if q_match and (not o_match or k < j):
            _push(b[i:k])
            _push(b[k + 1:k + 2])
            i = k + 2
        else:
            _push(b[i:j])
            rv.append(int(b[j + 1:j + 4], 8))
            i = j + 4

    return bytes(rv)


def _make_cookie_domain(domain):
    if domain is None:
        return None
    domain = encode_idna(domain)
    if b':' in domain:
        domain = domain.split(b':', 1)[0]
    if b'.' in domain:
        return domain
    raise ValueError(
        'Setting \'domain\' for a cookie on a server running locally (ex: '
        'localhost) is not supported by complying browsers. You should '
        'have something like: \'127.0.0.1 localhost dev.localhost\' on '
        'your hosts file and then point your server to run on '
        '\'dev.localhost\' and also set \'domain\' for \'dev.localhost\''
    )


def _cookie_parse_impl(b):
    """Lowlevel cookie parsing facility that operates on bytes."""
    i = 0
    n = len(b)

    while i < n:
        match = _cookie_re.search(b + b';', i)
        if not match:
            break

        key = match.group('key').strip()
        value = match.group('val')
        i = match.end(0)

        # Ignore parameters.  We have no interest in them.
        if key.lower() not in _cookie_params:
            yield _cookie_unquote(key), _cookie_unquote(value)


def cookie_date(expires=None):
    """Formats the time to ensure compatibility with Netscape's cookie
    standard.

    Accepts a floating point number expressed in seconds since the epoch in, a
    datetime object or a timetuple.  All times in UTC.  The :func:`parse_date`
    function can be used to parse such a date.

    Outputs a string in the format ``Wdy, DD-Mon-YYYY HH:MM:SS GMT``.

    :param expires:
        If provided that date is used, otherwise the current.
    """
    return _dump_date(expires, '-')


def parse_cookie(header, charset='utf-8', errors='replace', cls=None):
    """Parse a cookie.  Either from a string or WSGI environ.

    Per default encoding errors are ignored.  If you want a different behavior
    you can set `errors` to ``'replace'`` or ``'strict'``.  In strict mode a
    :exc:`HTTPUnicodeError` is raised.

    :param header:
        The header to be used to parse the cookie.  Alternatively this can be a
        WSGI environment.
    :param charset:
        The charset for the cookie values.
    :param errors:
        The error behavior for the charset decoding.
    :param cls:
        An optional dict class to use.  If this is not specified or `None` the
        default :class:`TypeConversionDict` is used.
    """
    if isinstance(header, dict):
        header = header.get('HTTP_COOKIE', '')
    elif header is None:
        header = ''

    # If the value is an unicode string it's mangled through latin1.  This
    # is done because on PEP 3333 on Python 3 all headers are assumed latin1
    # which however is incorrect for cookies, which are sent in page encoding.
    # As a result we
    if isinstance(header, str):
        header = header.encode('latin1', 'replace')

    if cls is None:
        cls = datastructures.TypeConversionDict

    def _parse_pairs():
        for key, val in _cookie_parse_impl(header):
            key = key.decode(charset)
            val = val.decode(charset)
            yield key, val

    return cls(_parse_pairs())


def dump_cookie(
    key, value='', max_age=None, expires=None, path='/',
    domain=None, secure=False, httponly=False,
    charset='utf-8', sync_expires=True,
):
    """Creates a new Set-Cookie header without the ``Set-Cookie`` prefix
    The parameters are the same as in the cookie Morsel object in the
    Python standard library but it accepts unicode data, too.

    The return value of this function will be a unicode string tunneled through
    latin1 as required by PEP 3333.

    The return value is not ASCII safe if the key contains unicode
    characters.  This is technically against the specification but
    happens in the wild.  It's strongly recommended to not use
    non-ASCII values for the keys.

    :param max_age:
        Should be a number of seconds, or `None` (default) if the cookie should
        last only as long as the client's browser session.  Additionally
        `timedelta` objects are accepted, too.
    :param expires:
        Should be a `datetime` object or unix timestamp.
    :param path:
        Limits the cookie to a given path, per default it will span the whole
        domain.
    :param domain:
        Use this if you want to set a cross-domain cookie. For example,
        ``domain=".example.com"`` will set a cookie that is readable by the
        domain ``www.example.com``, ``foo.example.com`` etc. Otherwise, a
        cookie will only be readable by the domain that set it.
    :param secure:
        The cookie will only be available via HTTPS.
    :param httponly:
        Disallow JavaScript to access the cookie.  This is an extension to the
        cookie standard and probably not supported by all browsers.
    :param charset:
        The encoding for unicode values.
    :param sync_expires:
        Automatically set expires if max_age is defined but expires not.
    """
    key = key.encode(charset)
    value = value.encode(charset)

    if path is not None:
        path = iri_to_uri(path)
    domain = _make_cookie_domain(domain)
    if isinstance(max_age, timedelta):
        max_age = (max_age.days * 60 * 60 * 24) + max_age.seconds
    if expires is not None:
        if not isinstance(expires, str):
            expires = cookie_date(expires)
    elif max_age is not None and sync_expires:
        expires = cookie_date(time() + max_age).encode('ascii')

    buf = [key + b'=' + _cookie_quote(value)]

    # XXX: In theory all of these parameters that are not marked with `None`
    # should be quoted.  Because stdlib did not quote it before I did not
    # want to introduce quoting there now.
    for k, v, q in (
        (b'Domain', domain, True),
        (b'Expires', expires, False,),
        (b'Max-Age', max_age, False),
        (b'Secure', secure, None),
        (b'HttpOnly', httponly, None),
        (b'Path', path, False),
    ):
        if q is None:
            if v:
                buf.append(k)
            continue

        if v is None:
            continue

        tmp = bytearray(k)
        if not isinstance(v, (bytes, bytearray)):
            v = str(v).encode(charset)
        if q:
            v = _cookie_quote(v)
        tmp += b'=' + v
        buf.append(bytes(tmp))

    # The return value will be an incorrectly encoded latin1 header on
    # for consistency with the headers object
    rv = b'; '.join(buf)
    rv = rv.decode('latin1')
    return rv



"""
    verktyg.views
    ~~~~~~~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import json

from verktyg.responses import Response
from verktyg.dispatch import BindingFactory, Binding


class View(BindingFactory):
    """ Wraps a function or callable so that it can be bound to a name in a
    dispatcher.

    :param name:
        The dispatcher name to bind to the action to.
    :param action:
        A callable object that takes an
        :class:`~verktyg.application.Application` and a
        :class:`~verktyg.requests.Request` and returns a
        :class:`~verktyg.responses.Response`
    :param methods:
        The http methods that this view can handle.  Can be a single method
        represented by a string or an iterable of method strings.
    :param content_type:
        The mime type of the response generated by the action
    :param language:
    :param charset:
    :param qs:
    """
    def __init__(
        self, name, action, *, methods=None,
        content_type=None, language=None, charset=None, qs=None
    ):
        self._name = name

        if methods is None:
            self._methods = {'GET'}
        elif isinstance(methods, str):
            self._methods = {methods}
        else:
            self._methods = set(methods)

        self._content_type = content_type
        self._language = language
        self._charset = charset
        self._qs = qs
        self._action = action

    def __call__(self, env, req, *args, **kwargs):
        return self._action(env, req, *args, **kwargs)

    def get_bindings(self):
        for method in self._methods:
            yield Binding(
                self._name, self, method=method,
                content_type=self._content_type, language=self._language,
                charset=self._charset, qs=self._qs,
            )


class ClassView(BindingFactory):
    def get_bindings(self):
        for method in {'GET', 'HEAD', 'POST', 'PUT', 'DELETE'}:  # TODO
            if hasattr(self, method):
                yield Binding(
                    self.name, getattr(self, method), method=method,
                )


class JsonView(View):
    def __init__(self, name, action, methods=None, qs=None):
        super(JsonView, self).__init__(
            name, action, methods=methods,
            content_type='application/json', qs=qs
        )

    def __call__(self, env, req, *args, **kwargs):
        res = super(JsonView, self).__call__(env, req, *args, **kwargs)

        if isinstance(res, Response):
            # rendering already done
            return res

        if res is None:
            # no content
            return Response(status=204)

        if env.debug:
            json_response = json.dumps(res, indent=4)
        else:
            json_response = json.dumps(res, separators=(',', ':'))

        return Response(json_response, content_type='application/json')


def expose(dispatcher, name, *args, **kwargs):
    def decorator(f):
        dispatcher.add_bindings(View(name, f, *args, **kwargs))
        return f
    return decorator


def expose_json(dispatcher, name, *args, **kwargs):
    def decorator(f):
        dispatcher.add_bindings(JsonView(name, f, *args, **kwargs))
        return f
    return decorator



"""
    tests.requests
    ~~~~~~~~~~~~~~

    Tests for request objects.

    :copyright:
        (c) 2017 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

import pickle
from io import BytesIO
from datetime import datetime

from verktyg.datastructures import (
    MultiDict, ImmutableList,
    ImmutableTypeConversionDict,
)
from verktyg.test import Client
from verktyg.wsgi import LimitedStream
from verktyg.exceptions import SecurityError
from verktyg.requests import BaseRequest, Request
from verktyg.responses import BaseResponse


class RequestTestResponse(BaseResponse):

    """Subclass of the normal response class we use to test response
    and base classes.  Has some methods to test if things in the
    response match.
    """

    def __init__(self, response, status, headers):
        BaseResponse.__init__(self, response, status, headers)
        self.body_data = pickle.loads(self.get_data())

    def __getitem__(self, key):
        return self.body_data[key]


def request_demo_app(environ, start_response):
    request = BaseRequest(environ)
    assert 'verktyg.request' in environ
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return [pickle.dumps({
        'args': request.args,
        'args_as_list': list(request.args.lists()),
        'environ': prepare_environ_pickle(request.environ),
        'data': request.get_data()
    })]


def prepare_environ_pickle(environ):
    result = {}
    for key, value in environ.items():
        try:
            pickle.dumps((key, value))
        except Exception:
            continue
        result[key] = value
    return result


class RequestsTestCase(unittest.TestCase):
    def assert_environ(self, environ, method):
        self.assertEqual(environ['REQUEST_METHOD'], method)
        self.assertEqual(environ['PATH_INFO'], '/')
        self.assertEqual(environ['SCRIPT_NAME'], '')
        self.assertEqual(environ['SERVER_NAME'], 'localhost')
        self.assertEqual(environ['wsgi.version'], (1, 0))
        self.assertEqual(environ['wsgi.url_scheme'], 'http')

    def test_base_request(self):
        client = Client(request_demo_app, RequestTestResponse)

        # get requests
        response = client.get('/?foo=bar&foo=hehe')
        self.assertEqual(
            response['args'], MultiDict([('foo', u'bar'), ('foo', u'hehe')])
        )
        self.assertEqual(
            response['args_as_list'], [('foo', [u'bar', u'hehe'])]
        )
        self.assertEqual(response['data'], b'')
        self.assert_environ(response['environ'], 'GET')

        # post requests with json data
        json = b'{"foo": "bar", "blub": "blah"}'
        response = client.post(
            '/?a=b', data=json, content_type='application/json'
        )
        self.assertEqual(response['data'], json)
        self.assertEqual(response['args'], MultiDict([('a', u'b')]))

    def test_query_string_is_bytes(self):
        req = Request.from_values(u'/?foo=%2f')
        self.assertEqual(req.query_string, 'foo=%2f')

    def test_request_repr(self):
        req = Request.from_values('/foobar')
        self.assertEqual(
            "<Request 'http://localhost/foobar' [GET]>", repr(req)
        )
        # test with non-ascii characters
        req = Request.from_values('/привет')
        self.assertEqual(
            "<Request 'http://localhost/привет' [GET]>", repr(req)
        )
        # test with unicode type for python 2
        req = Request.from_values(u'/привет')
        self.assertEqual(
            "<Request 'http://localhost/привет' [GET]>", repr(req)
        )

    def test_access_route(self):
        req = Request.from_values(headers={
            'X-Forwarded-For': '192.168.1.2, 192.168.1.1'
        })
        req.environ['REMOTE_ADDR'] = '192.168.1.3'
        self.assertEqual(req.access_route, ['192.168.1.2', '192.168.1.1'])
        self.assertEqual(req.remote_addr, '192.168.1.3')

        req = Request.from_values()
        req.environ['REMOTE_ADDR'] = '192.168.1.3'
        self.assertEqual(list(req.access_route), ['192.168.1.3'])

    def test_url_request_descriptors(self):
        req = Request.from_values(
            '/bar?foo=baz', 'http://example.com/test'
        )
        self.assertEqual(req.path, u'/bar')
        self.assertEqual(req.full_path, u'/bar?foo=baz')
        self.assertEqual(req.script_root, u'/test')
        self.assertEqual(req.url, u'http://example.com/test/bar?foo=baz')
        self.assertEqual(req.base_url, u'http://example.com/test/bar')
        self.assertEqual(req.url_root, u'http://example.com/test/')
        self.assertEqual(req.host_url, u'http://example.com/')
        self.assertEqual(req.host, 'example.com')
        self.assertEqual(req.scheme, 'http')

        req = Request.from_values(
            '/bar?foo=baz', 'https://example.com/test'
        )
        self.assertEqual(req.scheme, 'https')

    def test_url_request_descriptors_query_quoting(self):
        next = 'http%3A%2F%2Fwww.example.com%2F%3Fnext%3D%2F'
        req = Request.from_values(
            '/bar?next=' + next, 'http://example.com/'
        )
        self.assertEqual(req.path, u'/bar')
        self.assertEqual(req.full_path, u'/bar?next=' + next)
        self.assertEqual(req.url, u'http://example.com/bar?next=' + next)

    def test_url_request_descriptors_hosts(self):
        req = Request.from_values(
            '/bar?foo=baz', 'http://example.com/test'
        )
        req.trusted_hosts = ['example.com']
        self.assertEqual(req.path, u'/bar')
        self.assertEqual(req.full_path, u'/bar?foo=baz')
        self.assertEqual(req.script_root, u'/test')
        self.assertEqual(req.url, u'http://example.com/test/bar?foo=baz')
        self.assertEqual(req.base_url, u'http://example.com/test/bar')
        self.assertEqual(req.url_root, u'http://example.com/test/')
        self.assertEqual(req.host_url, u'http://example.com/')
        self.assertEqual(req.host, 'example.com')
        self.assertEqual(req.scheme, 'http')

        req = Request.from_values(
            '/bar?foo=baz', 'https://example.com/test'
        )
        self.assertEqual(req.scheme, 'https')

        req = Request.from_values(
            '/bar?foo=baz', 'http://example.com/test'
        )
        req.trusted_hosts = ['example.org']
        self.assertRaises(SecurityError, lambda: req.url)
        self.assertRaises(SecurityError, lambda: req.base_url)
        self.assertRaises(SecurityError, lambda: req.url_root)
        self.assertRaises(SecurityError, lambda: req.host_url)
        self.assertRaises(SecurityError, lambda: req.host)

    def test_authorization_mixin(self):
        request = Request.from_values(headers={
            'Authorization': 'Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ=='
        })
        a = request.authorization
        self.assertEqual(a.type, 'basic')
        self.assertEqual(a.username, 'Aladdin')
        self.assertEqual(a.password, 'open sesame')

    def test_etag_request_mixin(self):
        request = Request({
            'HTTP_CACHE_CONTROL': 'no-store, no-cache',
            'HTTP_IF_MATCH': 'w/"foo", bar, "baz"',
            'HTTP_IF_NONE_MATCH': 'w/"foo", bar, "baz"',
            'HTTP_IF_MODIFIED_SINCE': 'Tue, 22 Jan 2008 11:18:44 GMT',
            'HTTP_IF_UNMODIFIED_SINCE': 'Tue, 22 Jan 2008 11:18:44 GMT'
        })
        self.assertTrue(request.cache_control.no_store)
        self.assertTrue(request.cache_control.no_cache)

        for etags in request.if_match, request.if_none_match:
            self.assertTrue(etags('bar'))
            self.assertTrue(etags.contains_raw('w/"foo"'))
            self.assertTrue(etags.contains_weak('foo'))
            self.assertFalse(etags.contains('foo'))

        self.assertEqual(
            request.if_modified_since, datetime(2008, 1, 22, 11, 18, 44)
        )
        self.assertEqual(
            request.if_unmodified_since, datetime(2008, 1, 22, 11, 18, 44)
        )

    def test_common_request_descriptors_mixin(self):
        request = Request.from_values(
            content_type='text/html; charset=utf-8',
            content_length='23',
            headers={
                'Referer': 'http://www.example.com/',
                'Date': 'Sat, 28 Feb 2009 19:04:35 GMT',
                'Max-Forwards': '10',
                'Pragma': 'no-cache',
                'Content-Encoding': 'gzip',
                'Content-MD5': '9a3bc6dbc47a70db25b84c6e5867a072'
            }
        )

        self.assertEqual(request.content_type, 'text/html; charset=utf-8')
        self.assertEqual(request.mimetype, 'text/html')
        self.assertEqual(request.mimetype_params, {'charset': 'utf-8'})
        self.assertEqual(request.content_length, 23)
        self.assertEqual(request.referrer, 'http://www.example.com/')
        self.assertEqual(request.date, datetime(2009, 2, 28, 19, 4, 35))
        self.assertEqual(request.max_forwards, 10)
        self.assertTrue('no-cache' in request.pragma)
        self.assertEqual(request.content_encoding, 'gzip')
        self.assertEqual(
            request.content_md5, '9a3bc6dbc47a70db25b84c6e5867a072'
        )

    def test_request_mimetype_always_lowercase(self):
        request = Request.from_values(content_type='APPLICATION/JSON')
        self.assertEqual(request.mimetype, 'application/json')

    def test_shallow_mode(self):
        request = Request({'QUERY_STRING': 'foo=bar'}, shallow=True)
        self.assertEqual(request.args['foo'], 'bar')

    def test_other_method_payload(self):
        data = b'Hello World'
        req = Request.from_values(
            input_stream=BytesIO(data), method='WHAT_THE_FUCK',
            content_length=len(data), content_type='text/plain'
        )
        self.assertEqual(req.get_data(), data)
        self.assertTrue(isinstance(req.stream, LimitedStream))

    def test_storage_classes(self):
        class MyRequest(Request):
            dict_storage_class = dict
            list_storage_class = list
            parameter_storage_class = dict
        req = MyRequest.from_values('/?foo=baz', headers={
            'Cookie': 'foo=bar'
        })
        self.assertIs(type(req.cookies), dict)
        self.assertEqual(req.cookies, {'foo': 'bar'})
        self.assertIs(type(req.access_route), list)

        self.assertIs(type(req.args), dict)

        req = Request.from_values(headers={
            'Cookie': 'foo=bar'
        })
        self.assertIs(type(req.cookies), ImmutableTypeConversionDict)
        self.assertEqual(req.cookies, {'foo': 'bar'})
        self.assertIs(type(req.access_route), ImmutableList)

        MyRequest.list_storage_class = tuple
        req = MyRequest.from_values()
        self.assertIs(type(req.access_route), tuple)

    def test_request_method_case_sensitivity(self):
        req = Request({'REQUEST_METHOD': 'get'})
        self.assertEqual(req.method, 'GET')



"""
    verktyg.testsuite.accept
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Tests for choosing representations based on content type, etc.

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import unittest

from verktyg.exceptions import NotAcceptable
from verktyg.accept import (
    Representation, select_representation, parse_accept_header,
)


class RepresentationTestCase(unittest.TestCase):
    def test_content_type_dispatch(self):
        accept = parse_accept_header(
            'text/xml,'
            'application/xml,'
            'application/xhtml+xml,'
            'application/foo;q=0.6;quiet=no;bar=baz,'
            'text/html;q=0.9,'
            'text/plain;q=0.8,'
            'image/png,'
            '*/*;q=0.5'
        )

        default_repr = Representation()
        png_repr = Representation(content_type='image/png')
        plain_repr = Representation(content_type='text/plain')
        json_repr = Representation(content_type='application/json')

        default_match = default_repr.acceptability(accept=accept)
        png_match = png_repr.acceptability(accept=accept)
        plain_match = plain_repr.acceptability(accept=accept)
        json_match = json_repr.acceptability(accept=accept)

        self.assertLess(default_match, png_match)
        self.assertLess(default_match, plain_match)
        self.assertLess(default_match, json_match)
        self.assertGreater(png_match, plain_match)
        self.assertGreater(png_match, json_match)
        self.assertGreater(plain_match, json_match)

    def test_select_representation(self):
        json_repr = Representation(content_type='application/json')
        html_repr = Representation(content_type='text/html')
        xml_repr = Representation(content_type='application/xml')
        blank_repr = Representation()

        representations = [json_repr, html_repr, xml_repr, blank_repr]

        self.assertEqual(
            json_repr, select_representation(
                representations, accept='application/json'
            )
        )

        self.assertEqual(
            xml_repr,
            select_representation(
                representations,
                accept='application/xml;q=0.9,text/html;q=0.8'
            )
        )

        self.assertEqual(
            blank_repr,
            select_representation(
                representations, accept='image/png'
            )
        )

    def test_not_acceptable(self):
        pdf_repr = Representation(content_type='text/pdf')

        self.assertRaises(
            NotAcceptable,
            pdf_repr.acceptability,
            accept=parse_accept_header('text/html')
        )

        self.assertRaises(
            NotAcceptable,
            select_representation, [pdf_repr], accept='text/html'
        )



"""
    verktyg.testsuite.test_urls
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    URL helper tests.

    :copyright:
        (c) 2016 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

from verktyg import urls


class UrlsTestCase(unittest.TestCase):

    def test_url_fixing(self):
        x = urls.url_fix(
            'http://de.wikipedia.org/wiki/Elf (Begriffskl\xe4rung)'
        )
        self.assertEqual(
            x, 'http://de.wikipedia.org/wiki/Elf%20(Begriffskl%C3%A4rung)'
        )

        x = urls.url_fix("http://just.a.test/$-_.+!*'(),")
        self.assertEqual(x, "http://just.a.test/$-_.+!*'(),")

        x = urls.url_fix('http://höhöhö.at/höhöhö/hähähä')
        self.assertEqual(
            x,
            (
                r'http://xn--hhh-snabb.at'
                r'/h%C3%B6h%C3%B6h%C3%B6/h%C3%A4h%C3%A4h%C3%A4'
            )
        )

    def test_url_fixing_filepaths(self):
        x = urls.url_fix(r'file://C:\Users\Administrator\My Documents\ÑÈáÇíí')
        self.assertEqual(
            x,
            (
                r'file:///C%3A/Users/Administrator/My%20Documents/'
                r'%C3%91%C3%88%C3%A1%C3%87%C3%AD%C3%AD'
            )
        )

        a = urls.url_fix(r'file:/C:/')
        b = urls.url_fix(r'file://C:/')
        c = urls.url_fix(r'file:///C:/')
        self.assertEqual(a, r'file:///C%3A/')
        self.assertEqual(b, r'file:///C%3A/')
        self.assertEqual(c, r'file:///C%3A/')

        x = urls.url_fix(r'file://host/sub/path')
        self.assertEqual(x, r'file://host/sub/path')

        x = urls.url_fix(r'file:///')
        self.assertEqual(x, r'file:///')

    def test_url_fixing_qs(self):
        x = urls.url_fix('http://example.com/?foo=%2f%2f')
        self.assertEqual(x, 'http://example.com/?foo=%2f%2f')

        x = urls.url_fix(
            'http://acronyms.thefreedictionary.com/'
            'Algebraic+Methods+of+Solving+the+Schr%C3%B6dinger+Equation'
        )
        self.assertEqual(
            x,
            (
                'http://acronyms.thefreedictionary.com/'
                'Algebraic+Methods+of+Solving+the+Schr%C3%B6dinger+Equation'
            )
        )

    def test_iri_support(self):
        self.assertEqual(
            urls.uri_to_iri('http://xn--n3h.net/'),
            'http://\u2603.net/'
        )
        self.assertEqual(
            urls.uri_to_iri(
                'http://%C3%BCser:p%C3%A4ssword@xn--n3h.net/p%C3%A5th'
            ),
            'http://\xfcser:p\xe4ssword@\u2603.net/p\xe5th'
        )
        self.assertEqual(
            urls.iri_to_uri('http://☃.net/'),
            'http://xn--n3h.net/'
        )
        self.assertEqual(
            urls.iri_to_uri('http://üser:pässword@☃.net/påth'),
            'http://%C3%BCser:p%C3%A4ssword@xn--n3h.net/p%C3%A5th'
        )

        self.assertEqual(
            urls.uri_to_iri('http://test.com/%3Fmeh?foo=%26%2F'),
            'http://test.com/%3Fmeh?foo=%26%2F'
        )

        self.assertEqual(urls.iri_to_uri('/foo'), '/foo')

        self.assertEqual(
            urls.iri_to_uri('http://föö.com:8080/bam/baz'),
            'http://xn--f-1gaa.com:8080/bam/baz'
        )

    def test_iri_safe_quoting(self):
        uri = 'http://xn--f-1gaa.com/%2F%25?q=%C3%B6&x=%3D%25#%25'
        iri = 'http://föö.com/%2F%25?q=ö&x=%3D%25#%25'
        self.assertEqual(urls.uri_to_iri(uri), iri)
        self.assertEqual(urls.iri_to_uri(urls.uri_to_iri(uri)), uri)

    def test_iri_to_uri_idempotence_ascii_only(self):
        uri = 'http://www.idempoten.ce'
        uri = urls.iri_to_uri(uri)
        self.assertEqual(urls.iri_to_uri(uri), uri)

    def test_iri_to_uri_idempotence_non_ascii(self):
        uri = 'http://\N{SNOWMAN}/\N{SNOWMAN}'
        uri = urls.iri_to_uri(uri)
        self.assertEqual(urls.iri_to_uri(uri), uri)

    def test_uri_to_iri_idempotence_ascii_only(self):
        uri = 'http://www.idempoten.ce'
        uri = urls.uri_to_iri(uri)
        self.assertEqual(urls.uri_to_iri(uri), uri)

    def test_uri_to_iri_idempotence_non_ascii(self):
        uri = 'http://xn--n3h/%E2%98%83'
        uri = urls.uri_to_iri(uri)
        self.assertEqual(urls.uri_to_iri(uri), uri)

    def test_iri_to_uri_to_iri(self):
        iri = 'http://föö.com/'
        uri = urls.iri_to_uri(iri)
        self.assertEqual(urls.uri_to_iri(uri), iri)

    def test_uri_to_iri_to_uri(self):
        uri = 'http://xn--f-rgao.com/%C3%9E'
        iri = urls.uri_to_iri(uri)
        self.assertEqual(urls.iri_to_uri(iri), uri)

    def test_uri_iri_normalization(self):
        uri = 'http://xn--f-rgao.com/%E2%98%90/fred?utf8=%E2%9C%93'
        iri = 'http://föñ.com/\N{BALLOT BOX}/fred?utf8=\u2713'

        tests = [
            'http://föñ.com/\N{BALLOT BOX}/fred?utf8=\u2713',
            'http://xn--f-rgao.com/\u2610/fred?utf8=\N{CHECK MARK}',
            'http://xn--f-rgao.com/%E2%98%90/fred?utf8=%E2%9C%93',
            'http://xn--f-rgao.com/%E2%98%90/fred?utf8=%E2%9C%93',
            'http://föñ.com/\u2610/fred?utf8=%E2%9C%93',
        ]

        for test in tests:
            self.assertEqual(urls.uri_to_iri(test), iri)
            self.assertEqual(urls.iri_to_uri(test), uri)
            self.assertEqual(urls.uri_to_iri(urls.iri_to_uri(test)), iri)
            self.assertEqual(urls.iri_to_uri(urls.uri_to_iri(test)), uri)
            self.assertEqual(urls.uri_to_iri(urls.uri_to_iri(test)), iri)
            self.assertEqual(urls.iri_to_uri(urls.iri_to_uri(test)), uri)



"""
    verktyg.accept
    ~~~~~~~~~~~~~~

    Code for selecting handlers based on accept headers

    http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html

    :copyright: (c) 2015 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import functools

from verktyg.accept.content_type import (
    parse_content_type_header, parse_accept_header
)
from verktyg.accept.language import (
    parse_language_header, parse_accept_language_header,
)
from verktyg.accept.charset import (
    parse_charset_header, parse_accept_charset_header,
)

from verktyg.exceptions import NotAcceptable


@functools.total_ordering
class Acceptability(object):
    def __init__(
        self, *,
        content_type_acceptability,
        language_acceptability,
        charset_acceptability,
        qs
    ):
        self._content_type_acceptability = content_type_acceptability
        self._language_acceptability = language_acceptability
        self._charset_acceptability = charset_acceptability
        self._qs = qs

    def __eq__(self, other):
        if (
            other._content_type_acceptability !=
            self._content_type_acceptability
        ):
            return False
        if other._language_acceptability != self._language_acceptability:
            return False
        if other._charset_acceptability != self._charset_acceptability:
            return False
        if other._qs != self._qs:
            return False
        return True

    def __gt__(self, other):
        def _gt(self, other):
            if (self is None) and (other is None):
                return False
            if self <= other:
                return False
            return True

        if _gt(
                self._content_type_acceptability,
                other._content_type_acceptability):
            return True
        if _gt(
                self._language_acceptability,
                other._language_acceptability):
            return True
        if _gt(
                self._charset_acceptability,
                other._charset_acceptability):
            return True
        if _gt(
                self._qs,
                other._qs):
            return True
        return False


def select_representation(
    representations,
    accept='*/*', accept_language='*', accept_charset='*'
):
    highest_acceptability = None
    best_representation = None

    if accept is None:
        accept = '*/*'
    if isinstance(accept, str):
        accept = parse_accept_header(accept)

    if accept_language is None:
        accept_language = '*'
    if isinstance(accept_language, str):
        accept_language = parse_accept_language_header(accept_language)

    if accept_charset is None:
        accept_charset = '*'
    if isinstance(accept_charset, str):
        accept_charset = parse_accept_charset_header(accept_charset)

    for representation in representations:
        try:
            acceptability = representation.acceptability(
                accept=accept,
                accept_language=accept_language,
                accept_charset=accept_charset
            )
        except NotAcceptable:
            continue

        if (highest_acceptability is None or
                acceptability >= highest_acceptability):
            highest_acceptability = acceptability
            best_representation = representation

    if best_representation is None:
        raise NotAcceptable()

    return best_representation


class Representation(object):
    def __init__(
        self, *, content_type=None, language=None,
        charset=None, qs=None
    ):
        if isinstance(content_type, str):
            content_type = parse_content_type_header(content_type)
        self._content_type = content_type

        if isinstance(language, str):
            language = parse_language_header(language)
        self._language = language

        if isinstance(charset, str):
            charset = parse_charset_header(charset)
        self._charset = charset

        if qs is None:
            qs = 1.0
        self._qs = max(min(float(qs), 1.0), 0.0)

    def acceptability(
        self, *,
        accept=None,
        accept_charset=None,
        accept_language=None
    ):
        """
        :param accept:
            String in the same format as an http `Accept` header

        :param accept_language:
            String in the same format as an http `Accept-Language` header

        :param accept_charset:
            String in the same format as an http `Accept-Charset` header

        :return:
            An orderable `Acceptability` object representing the quality of the
            match.

        :raises NotAcceptable: If the binding does not match the request.
        """
        content_type_acceptability = (
            self._content_type.acceptability(accept)
            if self._content_type is not None else
            None
        )

        language_acceptability = (
            self._content_type.acceptability(accept)
            if self._language is not None else
            None
        )

        charset_acceptability = (
            self._charset.acceptability
            if self._charset is not None else
            None
        )

        return Acceptability(
            content_type_acceptability=content_type_acceptability,
            language_acceptability=language_acceptability,
            charset_acceptability=charset_acceptability,
            qs=self._qs,
        )

    def __repr__(self):
        output = "<%s" % self.__class__.__name__
        if self._content_type is not None:
            output += " content_type=%r" % self._content_type
        if self._language is not None:
            output += " language=%r" % self._language
        if self._charset is not None:
            output += " charset=%r" % self._charset
        output += ">"
        return output



"""
    tests.exceptions
    ~~~~~~~~~~~~~~~~

    The tests for the exception classes.

    :copyright: (c) 2014 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.
"""
import unittest

from verktyg import exceptions


class ExceptionsTestCase(unittest.TestCase):
    def test_aborter_general(self):
        codes = [
            (exceptions.BadRequest, 400),
            (exceptions.Unauthorized, 401),
            (exceptions.Forbidden, 403),
            (exceptions.NotFound, 404),
            (exceptions.MethodNotAllowed, 405, ['GET', 'HEAD']),
            (exceptions.NotAcceptable, 406),
            (exceptions.RequestTimeout, 408),
            (exceptions.Gone, 410),
            (exceptions.LengthRequired, 411),
            (exceptions.PreconditionFailed, 412),
            (exceptions.RequestEntityTooLarge, 413),
            (exceptions.RequestURITooLarge, 414),
            (exceptions.UnsupportedMediaType, 415),
            (exceptions.UnprocessableEntity, 422),
            (exceptions.InternalServerError, 500),
            (exceptions.NotImplemented, 501),
            (exceptions.BadGateway, 502),
            (exceptions.ServiceUnavailable, 503)
        ]
        for exc_type, code, *args in codes:
            with self.subTest(code=code):
                self.assertRaises(exc_type, exceptions.abort, code, *args)

    def test_aborter_custom(self):
        myabort = exceptions.Aborter({1: exceptions.NotFound})
        self.assertRaises(LookupError, myabort, 404)
        self.assertRaises(exceptions.NotFound, myabort, 1)

        myabort = exceptions.Aborter(extra={1: exceptions.NotFound})
        self.assertRaises(exceptions.NotFound, myabort, 404)
        self.assertRaises(exceptions.NotFound, myabort, 1)

    def test_exception_repr(self):
        exc = exceptions.NotFound()
        self.assertEqual(str(exc), '404: Not Found')
        self.assertEqual(repr(exc), "<NotFound '404: Not Found'>")



"""
    tests.responses
    ~~~~~~~~~~~~~~

    Tests for response objects.

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

from datetime import datetime

from verktyg.http import Headers
from verktyg.test import create_environ, run_wsgi_app
from verktyg.requests import Request
from verktyg.responses import (
    BaseResponse, Response, ETagResponseMixin, generate_etag
)


class ResponsesTestCase(unittest.TestCase):
    def assert_environ(self, environ, method):
        self.assertEqual(environ['REQUEST_METHOD'], method)
        self.assertEqual(environ['PATH_INFO'], '/')
        self.assertEqual(environ['SCRIPT_NAME'], '')
        self.assertEqual(environ['SERVER_NAME'], 'localhost')
        self.assertEqual(environ['wsgi.version'], (1, 0))
        self.assertEqual(environ['wsgi.url_scheme'], 'http')

    def test_base_response(self):
        # unicode
        response = BaseResponse(u'öäü')
        self.assertEqual(response.get_data(), u'öäü'.encode('utf-8'))

        # writing
        response = Response('foo')
        response.stream.write('bar')
        self.assertEqual(response.get_data(), b'foobar')

        # set cookie
        response = BaseResponse()
        response.set_cookie('foo', 'bar', 60, 0, '/blub', 'example.org')
        self.assertEqual(response.headers.to_wsgi_list(), [
            ('Content-Type', 'text/plain; charset=utf-8'),
            ('Set-Cookie', (
                'foo=bar; '
                'Domain=example.org; '
                'Expires=Thu, 01-Jan-1970 00:00:00 GMT; '
                'Max-Age=60; '
                'Path=/blub'
            )),
        ])

        # delete cookie
        response = BaseResponse()
        response.delete_cookie('foo')
        self.assertEqual(response.headers.to_wsgi_list(), [
            ('Content-Type', 'text/plain; charset=utf-8'),
            ('Set-Cookie', (
                'foo=; '
                'Expires=Thu, 01-Jan-1970 00:00:00 GMT; '
                'Max-Age=0; '
                'Path=/'
            )),
        ])

        # close call forwarding
        closed = []

        class Iterable(object):

            def __next__(self):
                raise StopIteration()

            def __iter__(self):
                return self

            def close(self):
                closed.append(True)
        response = BaseResponse(Iterable())
        response.call_on_close(lambda: closed.append(True))
        app_iter, status, headers = run_wsgi_app(
            response, create_environ(), buffered=True,
        )
        self.assertEqual(status, '200 OK')
        self.assertEqual(''.join(app_iter), '')
        self.assertEqual(len(closed), 2)

        # with statement
        del closed[:]
        response = BaseResponse(Iterable())
        with response:
            pass
        self.assertEqual(len(closed), 1)

    def test_response_status_codes(self):
        response = BaseResponse()
        response.status_code = 404
        self.assertEqual(response.status, '404 NOT FOUND')
        response.status = '200 OK'
        self.assertEqual(response.status_code, 200)
        response.status = '999 WTF'
        self.assertEqual(response.status_code, 999)
        response.status_code = 588
        self.assertEqual(response.status_code, 588)
        self.assertEqual(response.status, '588 UNKNOWN')
        response.status = 'wtf'
        self.assertEqual(response.status_code, 0)
        self.assertEqual(response.status, '0 wtf')

    def test_type_forcing(self):
        def wsgi_application(environ, start_response):
            start_response('200 OK', [('Content-Type', 'text/html')])
            return ['Hello World!']
        base_response = BaseResponse(
            'Hello World!', content_type='text/html'
        )

        class SpecialResponse(Response):

            def foo(self):
                return 42

        # good enough for this simple application, but don't ever use that in
        # real world examples!
        fake_env = {}

        for orig_resp in wsgi_application, base_response:
            response = SpecialResponse.force_type(orig_resp, fake_env)
            self.assertEqual(response.__class__, SpecialResponse)
            self.assertEqual(response.foo(), 42)
            self.assertEqual(response.get_data(), b'Hello World!')
            self.assertEqual(response.content_type, 'text/html')

        # without env, no arbitrary conversion
        self.assertRaises(
            TypeError, SpecialResponse.force_type, wsgi_application
        )

    def test_etag_response_mixin(self):
        response = Response('Hello World')
        self.assertEqual(response.get_etag(), (None, None))
        response.add_etag()
        self.assertEqual(
            response.get_etag(), ('b10a8db164e0754105b7a99be72e3fe5', False)
        )
        self.assertFalse(response.cache_control)
        response.cache_control.must_revalidate = True
        response.cache_control.max_age = 60
        response.headers['Content-Length'] = len(response.get_data())
        self.assertIn(
            response.headers['Cache-Control'],
            ('must-revalidate, max-age=60', 'max-age=60, must-revalidate')
        )

        self.assertNotIn('date', response.headers)
        env = create_environ()
        env.update({
            'REQUEST_METHOD': 'GET',
            'HTTP_IF_NONE_MATCH': response.get_etag()[0]
        })
        response.make_conditional(env)
        self.assertIn('date', response.headers)

        # after the thing is invoked by the server as wsgi application
        # (we're emulating this here), there must not be any entity
        # headers left and the status code would have to be 304
        resp = Response.from_app(response, env)
        self.assertEqual(resp.status_code, 304)
        self.assertNotIn('content-length', resp.headers)

        # make sure date is not overriden
        response = Response('Hello World')
        response.date = 1337
        d = response.date
        response.make_conditional(env)
        self.assertEqual(response.date, d)

        # make sure content length is only set if missing
        response = Response('Hello World')
        response.content_length = 999
        response.make_conditional(env)
        self.assertEqual(response.content_length, 999)

    def test_etag_response_mixin_freezing(self):
        class WithFreeze(ETagResponseMixin, BaseResponse):
            pass

        class WithoutFreeze(BaseResponse, ETagResponseMixin):
            pass

        response = WithFreeze('Hello World')
        response.freeze()
        self.assertEqual(
            response.get_etag(),
            (str(generate_etag(b'Hello World')), False)
        )
        response = WithoutFreeze('Hello World')
        response.freeze()
        self.assertEqual(response.get_etag(), (None, None))
        response = Response('Hello World')
        response.freeze()
        self.assertEqual(response.get_etag(), (None, None))

    def test_authenticate_mixin(self):
        resp = Response()
        resp.www_authenticate.type = 'basic'
        resp.www_authenticate.realm = 'Testing'
        self.assertEqual(
            resp.headers['WWW-Authenticate'], u'Basic realm="Testing"'
        )
        resp.www_authenticate.realm = None
        resp.www_authenticate.type = None
        self.assertNotIn('WWW-Authenticate', resp.headers)

    def test_authenticate_mixin_quoted_qop(self):
        # Example taken from https://github.com/mitsuhiko/werkzeug/issues/633
        resp = Response()
        resp.www_authenticate.set_digest(
            'REALM', 'NONCE', qop=("auth", "auth-int")
        )

        actual = set((resp.headers['WWW-Authenticate'] + ',').split())
        expected = set((
            'Digest nonce="NONCE", realm="REALM", qop="auth, auth-int",'
        ).split())
        self.assertEqual(actual, expected)

        resp.www_authenticate.set_digest('REALM', 'NONCE', qop=("auth",))

        actual = set((resp.headers['WWW-Authenticate'] + ',').split())
        expected = set(
            'Digest nonce="NONCE", realm="REALM", qop="auth",'.split()
        )
        self.assertEqual(actual, expected)

    def test_response_stream_mixin(self):
        response = Response()
        response.stream.write('Hello ')
        response.stream.write('World!')
        self.assertEqual(response.response, ['Hello ', 'World!'])
        self.assertEqual(response.get_data(), b'Hello World!')

    def test_common_response_descriptors_mixin(self):
        response = Response()
        response.mimetype = 'text/html'
        self.assertEqual(response.mimetype, 'text/html')
        self.assertEqual(response.content_type, 'text/html; charset=utf-8')
        self.assertEqual(response.mimetype_params, {'charset': 'utf-8'})
        response.mimetype_params['x-foo'] = 'yep'
        del response.mimetype_params['charset']
        self.assertEqual(response.content_type, 'text/html; x-foo=yep')

        now = datetime.utcnow().replace(microsecond=0)

        self.assertIsNone(response.content_length)
        response.content_length = '42'
        self.assertEqual(response.content_length, 42)

        for attr in 'date', 'age', 'expires':
            self.assertIsNone(getattr(response, attr))
            setattr(response, attr, now)
            self.assertEqual(getattr(response, attr), now)

        self.assertIsNone(response.retry_after)
        response.retry_after = now
        self.assertEqual(response.retry_after, now)

        self.assertFalse(response.vary)
        response.vary.add('Cookie')
        response.vary.add('Content-Language')
        self.assertTrue('cookie' in response.vary)
        self.assertEqual(response.vary.to_header(), 'Cookie, Content-Language')
        response.headers['Vary'] = 'Content-Encoding'
        self.assertEqual(response.vary.as_set(), set(['content-encoding']))

        response.allow.update(['GET', 'POST'])
        self.assertEqual(response.headers['Allow'], 'GET, POST')

        response.content_language.add('en-US')
        response.content_language.add('fr')
        self.assertEqual(response.headers['Content-Language'], 'en-US, fr')

    def test_response_streamed(self):
        r = Response()
        self.assertFalse(r.is_streamed)
        r = Response("Hello World")
        self.assertFalse(r.is_streamed)
        r = Response(["foo", "bar"])
        self.assertFalse(r.is_streamed)

        def gen():
            if 0:
                yield None
        r = Response(gen())
        self.assertTrue(r.is_streamed)

    def test_response_iter_wrapping(self):
        def uppercasing(iterator):
            for item in iterator:
                yield item.upper()

        def generator():
            yield 'foo'
            yield 'bar'
        req = Request.from_values()
        resp = Response(generator())
        del resp.headers['Content-Length']
        resp.response = uppercasing(resp.iter_encoded())
        actual_resp = Response.from_app(
            resp, req.environ, buffered=True
        )
        self.assertEqual(actual_resp.get_data(), b'FOOBAR')

    def test_response_freeze(self):
        def generate():
            yield "foo"
            yield "bar"
        resp = Response(generate())
        resp.freeze()
        self.assertEqual(resp.response, [b'foo', b'bar'])
        self.assertEqual(resp.headers['content-length'], '6')

    def test_urlfication(self):
        resp = Response()
        resp.headers['Location'] = u'http://üser:pässword@☃.net/påth'
        resp.headers['Content-Location'] = u'http://☃.net/'
        headers = resp.get_wsgi_headers(create_environ())
        self.assertEqual(
            headers['location'],
            'http://%C3%BCser:p%C3%A4ssword@xn--n3h.net/p%C3%A5th'
        )
        self.assertEqual(
            headers['content-location'],
            'http://xn--n3h.net/'
        )

    def test_new_response_iterator_behavior(self):
        req = Request.from_values()
        resp = Response(u'Hello Wörld!')

        def get_content_length(resp):
            headers = resp.get_wsgi_headers(req.environ)
            return headers.get('content-length', type=int)

        def generate_items():
            yield "Hello "
            yield u"Wörld!"

        # verktyg encodes when set to `data` now, which happens
        # if a string is passed to the response object.
        self.assertEqual(resp.response, [u'Hello Wörld!'.encode('utf-8')])
        self.assertEqual(resp.get_data(), u'Hello Wörld!'.encode('utf-8'))
        self.assertEqual(get_content_length(resp), 13)
        self.assertFalse(resp.is_streamed)
        self.assertTrue(resp.is_sequence)

        # try the same for manual assignment
        resp.set_data(u'Wörd')
        self.assertEqual(resp.response, [u'Wörd'.encode('utf-8')])
        self.assertEqual(resp.get_data(), u'Wörd'.encode('utf-8'))
        self.assertEqual(get_content_length(resp), 5)
        self.assertFalse(resp.is_streamed)
        self.assertTrue(resp.is_sequence)

        # automatic generator sequence conversion
        resp.response = generate_items()
        self.assertTrue(resp.is_streamed)
        self.assertFalse(resp.is_sequence)
        self.assertEqual(resp.get_data(), u'Hello Wörld!'.encode('utf-8'))
        self.assertEqual(resp.response, [b'Hello ', u'Wörld!'.encode('utf-8')])
        self.assertFalse(resp.is_streamed)
        self.assertTrue(resp.is_sequence)

        # automatic generator sequence conversion
        resp.response = generate_items()
        resp.implicit_sequence_conversion = False
        self.assertTrue(resp.is_streamed)
        self.assertFalse(resp.is_sequence)
        self.assertRaises(RuntimeError, lambda: resp.get_data())
        resp.make_sequence()
        self.assertEqual(resp.get_data(), u'Hello Wörld!'.encode('utf-8'))
        self.assertEqual(resp.response, [b'Hello ', u'Wörld!'.encode('utf-8')])
        self.assertFalse(resp.is_streamed)
        self.assertTrue(resp.is_sequence)

        # stream makes it a list no matter how the conversion is set
        for val in True, False:
            resp.implicit_sequence_conversion = val
            resp.response = ("foo", "bar")
            self.assertTrue(resp.is_sequence)
            resp.stream.write('baz')
            self.assertEqual(resp.response, ['foo', 'bar', 'baz'])

    def test_response_headers_passthrough(self):
        headers = Headers()
        resp = Response(headers=headers)
        self.assertIs(resp.headers, headers)

    def test_response_304_no_content_length(self):
        resp = Response('Test', status=304)
        env = create_environ()
        self.assertNotIn('content-length', resp.get_wsgi_headers(env))

    def test_ranges(self):
        # basic range stuff
        req = Request.from_values()
        self.assertIsNone(req.range)
        req = Request.from_values(headers={'Range': 'bytes=0-499'})
        self.assertEqual(req.range.ranges, [(0, 500)])

        resp = Response()
        resp.content_range = req.range.make_content_range(1000)
        self.assertEqual(resp.content_range.units, 'bytes')
        self.assertEqual(resp.content_range.start, 0)
        self.assertEqual(resp.content_range.stop, 500)
        self.assertEqual(resp.content_range.length, 1000)
        self.assertEqual(resp.headers['Content-Range'], 'bytes 0-499/1000')

        resp.content_range.unset()
        self.assertNotIn('Content-Range', resp.headers)

        resp.headers['Content-Range'] = 'bytes 0-499/1000'
        self.assertEqual(resp.content_range.units, 'bytes')
        self.assertEqual(resp.content_range.start, 0)
        self.assertEqual(resp.content_range.stop, 500)
        self.assertEqual(resp.content_range.length, 1000)

    def test_auto_content_length(self):
        resp = Response('Hello World!')
        self.assertEqual(resp.content_length, 12)

        resp = Response(['Hello World!'])
        self.assertIsNone(resp.content_length)
        self.assertEqual(resp.get_wsgi_headers({})['Content-Length'], '12')

    def test_stream_content_length(self):
        resp = Response()
        resp.stream.writelines(['foo', 'bar', 'baz'])
        self.assertEqual(resp.get_wsgi_headers({})['Content-Length'], '9')

        resp = Response()
        resp.make_conditional({'REQUEST_METHOD': 'GET'})
        resp.stream.writelines(['foo', 'bar', 'baz'])
        self.assertEqual(resp.get_wsgi_headers({})['Content-Length'], '9')

        resp = Response('foo')
        resp.stream.writelines(['bar', 'baz'])
        self.assertEqual(resp.get_wsgi_headers({})['Content-Length'], '9')

    def test_disabled_auto_content_length(self):
        class MyResponse(Response):
            automatically_set_content_length = False
        resp = MyResponse('Hello World!')
        self.assertIsNone(resp.content_length)

        resp = MyResponse(['Hello World!'])
        self.assertIsNone(resp.content_length)
        self.assertNotIn('Content-Length', resp.get_wsgi_headers({}))

        resp = MyResponse()
        resp.make_conditional({
            'REQUEST_METHOD': 'GET'
        })
        self.assertIsNone(resp.content_length)
        self.assertNotIn('Content-Length', resp.get_wsgi_headers({}))

    def test_location_header_autocorrect(self):
        env = create_environ()

        class MyResponse(Response):
            autocorrect_location_header = False
        resp = MyResponse('Hello World!')
        resp.headers['Location'] = '/test'
        self.assertEqual(resp.get_wsgi_headers(env)['Location'], '/test')

        resp = Response('Hello World!')
        resp.headers['Location'] = '/test'
        self.assertEqual(
            resp.get_wsgi_headers(env)['Location'], 'http://localhost/test'
        )



"""
    verktyg.http.auth
    ~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import base64

from verktyg import datastructures
from verktyg.http.basic import (
    wsgi_to_bytes, bytes_to_wsgi, parse_dict_header,
    dump_header, quote_header_value, parse_set_header,
)


class Authorization(datastructures.ImmutableDictMixin, dict):
    """Represents an `Authorization` header sent by the client.  You should
    not create this kind of object yourself but use it when it's returned by
    the `parse_authorization_header` function.

    This object is a dict subclass and can be altered by setting dict items
    but it should be considered immutable as it's returned by the client and
    not meant for modifications.
    """

    def __init__(self, auth_type, data=None):
        dict.__init__(self, data or {})
        self.type = auth_type

    username = property(lambda x: x.get('username'), doc='''
        The username transmitted.  This is set for both basic and digest
        auth all the time.''')
    password = property(lambda x: x.get('password'), doc='''
        When the authentication type is basic this is the password
        transmitted by the client, else `None`.''')
    realm = property(lambda x: x.get('realm'), doc='''
        This is the server realm sent back for HTTP digest auth.''')
    nonce = property(lambda x: x.get('nonce'), doc='''
        The nonce the server sent for digest auth, sent back by the client.
        A nonce should be unique for every 401 response for HTTP digest
        auth.''')
    uri = property(lambda x: x.get('uri'), doc='''
        The URI from Request-URI of the Request-Line; duplicated because
        proxies are allowed to change the Request-Line in transit.  HTTP
        digest auth only.''')
    nc = property(lambda x: x.get('nc'), doc='''
        The nonce count value transmitted by clients if a qop-header is
        also transmitted.  HTTP digest auth only.''')
    cnonce = property(lambda x: x.get('cnonce'), doc='''
        If the server sent a qop-header in the ``WWW-Authenticate``
        header, the client has to provide this value for HTTP digest auth.
        See the RFC for more details.''')
    response = property(lambda x: x.get('response'), doc='''
        A string of 32 hex digits computed as defined in RFC 2617, which
        proves that the user knows a password.  Digest auth only.''')
    opaque = property(lambda x: x.get('opaque'), doc='''
        The opaque header from the server returned unchanged by the client.
        It is recommended that this string be base64 or hexadecimal data.
        Digest auth only.''')

    @property
    def qop(self):
        """Indicates what "quality of protection" the client has applied to
        the message for HTTP digest auth."""
        def on_update(header_set):
            if not header_set and 'qop' in self:
                del self['qop']
            elif header_set:
                self['qop'] = header_set.to_header()
        return parse_set_header(self.get('qop'), on_update)


def parse_authorization_header(value):
    """Parse an HTTP basic/digest authorization header transmitted by the web
    browser.  The return value is either `None` if the header was invalid or
    not given, otherwise an :class:`~Authorization`
    object.

    :param value:
        The authorization header to parse.
    :return:
        A:class:`~Authorization` object or `None`.
    """
    if not value:
        return
    value = wsgi_to_bytes(value)
    try:
        auth_type, auth_info = value.split(None, 1)
        auth_type = auth_type.lower()
    except ValueError:
        return
    if auth_type == b'basic':
        try:
            username, password = base64.b64decode(auth_info).split(b':', 1)
        except Exception:
            return
        return Authorization('basic', {
            'username': bytes_to_wsgi(username),
            'password': bytes_to_wsgi(password),
        })
    elif auth_type == b'digest':
        auth_map = parse_dict_header(auth_info)
        for key in 'username', 'realm', 'nonce', 'uri', 'response':
            if key not in auth_map:
                return
        if 'qop' in auth_map:
            if not auth_map.get('nc') or not auth_map.get('cnonce'):
                return
        return Authorization('digest', auth_map)


def auth_property(name, doc=None):
    """A static helper function for subclasses to add extra authentication
    system properties onto a class::

        class FooAuthenticate(WWWAuthenticate):
            special_realm = auth_property('special_realm')

    For more information have a look at the sourcecode to see how the
    regular properties (:attr:`realm` etc.) are implemented.
    """

    def _set_value(self, value):
        if value is None:
            self.pop(name, None)
        else:
            self[name] = str(value)
    return property(lambda x: x.get(name), _set_value, doc=doc)


class WWWAuthenticate(datastructures.UpdateDictMixin, dict):
    """Provides simple access to `WWW-Authenticate` headers."""

    #: list of keys that require quoting in the generated header
    _require_quoting = frozenset(['domain', 'nonce', 'opaque', 'realm', 'qop'])

    def __init__(self, auth_type=None, values=None, on_update=None):
        dict.__init__(self, values or ())
        if auth_type:
            self['__auth_type__'] = auth_type
        self.on_update = on_update

    def set_basic(self, realm='authentication required'):
        """Clear the auth info and enable basic auth."""
        dict.clear(self)
        dict.update(self, {'__auth_type__': 'basic', 'realm': realm})
        if self.on_update:
            self.on_update(self)

    def set_digest(
        self, realm, nonce, qop=('auth',), opaque=None,
        algorithm=None, stale=False,
    ):
        """Clear the auth info and enable digest auth."""
        d = {
            '__auth_type__': 'digest',
            'realm': realm,
            'nonce': nonce,
            'qop': dump_header(qop)
        }
        if stale:
            d['stale'] = 'TRUE'
        if opaque is not None:
            d['opaque'] = opaque
        if algorithm is not None:
            d['algorithm'] = algorithm
        dict.clear(self)
        dict.update(self, d)
        if self.on_update:
            self.on_update(self)

    def to_header(self):
        """Convert the stored values into a WWW-Authenticate header."""
        d = dict(self)
        auth_type = d.pop('__auth_type__', None) or 'basic'
        return '%s %s' % (auth_type.title(), ', '.join([
            '%s=%s' % (key, quote_header_value(
                value, allow_token=key not in self._require_quoting
            ))
            for key, value in d.items()
        ]))

    def __str__(self):
        return self.to_header()

    def __repr__(self):
        return '<%s %r>' % (
            self.__class__.__name__,
            self.to_header()
        )

    def _set_property(name, doc=None):
        def fget(self):
            def on_update(header_set):
                if not header_set and name in self:
                    del self[name]
                elif header_set:
                    self[name] = header_set.to_header()
            return parse_set_header(self.get(name), on_update)
        return property(fget, doc=doc)

    type = auth_property('__auth_type__', doc='''
        The type of the auth mechanism.  HTTP currently specifies
        `Basic` and `Digest`.''')
    realm = auth_property('realm', doc='''
        A string to be displayed to users so they know which username and
        password to use.  This string should contain at least the name of
        the host performing the authentication and might additionally
        indicate the collection of users who might have access.''')
    domain = _set_property('domain', doc='''
        A list of URIs that define the protection space.  If a URI is an
        absolute path, it is relative to the canonical root URL of the
        server being accessed.''')
    nonce = auth_property('nonce', doc='''
        A server-specified data string which should be uniquely generated
        each time a 401 response is made.  It is recommended that this
        string be base64 or hexadecimal data.''')
    opaque = auth_property('opaque', doc='''
        A string of data, specified by the server, which should be returned
        by the client unchanged in the Authorization header of subsequent
        requests with URIs in the same protection space.  It is recommended
        that this string be base64 or hexadecimal data.''')
    algorithm = auth_property('algorithm', doc='''
        A string indicating a pair of algorithms used to produce the digest
        and a checksum.  If this is not present it is assumed to be "MD5".
        If the algorithm is not understood, the challenge should be ignored
        (and a different one used, if there is more than one).''')
    qop = _set_property('qop', doc='''
        A set of quality-of-privacy directives such as auth and auth-int.''')

    def _get_stale(self):
        val = self.get('stale')
        if val is not None:
            return val.lower() == 'true'

    def _set_stale(self, value):
        if value is None:
            self.pop('stale', None)
        else:
            self['stale'] = value and 'TRUE' or 'FALSE'
    stale = property(_get_stale, _set_stale, doc='''
        A flag, indicating that the previous request from the client was
        rejected because the nonce value was stale.''')
    del _get_stale, _set_stale
    del _set_property


# make auth_property a staticmethod so that subclasses of
# `WWWAuthenticate` can use it for new properties.
WWWAuthenticate.auth_property = staticmethod(auth_property)


def parse_www_authenticate_header(value, on_update=None):
    """Parse an HTTP WWW-Authenticate header into a
    :class:`~WWWAuthenticate` object.

    :param value:
        A WWW-Authenticate header to parse.
    :param on_update:
        An optional callable that is called every time a value on the
        :class:`~WWWAuthenticate` object is changed.
    :return:
        A:class:`~WWWAuthenticate` object.
    """
    if not value:
        return WWWAuthenticate(on_update=on_update)
    try:
        auth_type, auth_info = value.split(None, 1)
        auth_type = auth_type.lower()
    except (ValueError, AttributeError):
        return WWWAuthenticate(value.strip().lower(), on_update=on_update)
    return WWWAuthenticate(
        auth_type, parse_dict_header(auth_info), on_update
    )



"""
    verktyg.testsuite.application
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests `Application` utility class.

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import unittest

from verktyg.test import Client
from verktyg.exceptions import HTTPException, NotFound, ImATeapot
from verktyg.responses import Response, BaseResponse
from verktyg.views import expose
from verktyg.routing import Route
from verktyg.application import ApplicationBuilder


class ApplicationTestCase(unittest.TestCase):
    def test_basic(self):
        builder = ApplicationBuilder()

        builder.add_routes(Route('/', endpoint='index'))

        @expose(builder, 'index')
        def index(app, request):
            return Response('Hello World')

        app = builder()
        client = Client(app, BaseResponse)

        resp = client.get('/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.get_data(), b'Hello World')

    def test_exception_handlers(self):
        builder = ApplicationBuilder()

        @builder.exception_handler(BaseException)
        def default_handler(app, req, exc_type, exc_value, exc_traceback):
            return Response('default handler', status=500)

        @builder.exception_handler(HTTPException)
        def verktyg_handler(app, req, exc_type, exc_value, exc_traceback):
            return Response('verktyg handler', exc_value.code)

        @builder.expose(route='/raise_execption')
        def raise_exception(app, req):
            raise Exception()

        @builder.expose(route='/raise_key_error')
        def raise_key_error(app, req):
            raise KeyError()

        @builder.expose(route='/raise_teapot')
        def raise_teapot(app, req):
            raise ImATeapot()

        app = builder()
        client = Client(app, BaseResponse)

        resp = client.get('/raise_execption')
        self.assertEqual(resp.status_code, 500)
        self.assertEqual(resp.get_data(), b'default handler')

        resp = client.get('/raise_key_error')
        self.assertEqual(resp.status_code, 500)
        self.assertEqual(resp.get_data(), b'default handler')

        resp = client.get('/raise_teapot')
        self.assertEqual(resp.status_code, 418)
        self.assertEqual(resp.get_data(), b'verktyg handler')

    def test_middleware(self):
        builder = ApplicationBuilder()

        @builder.expose(route='/')
        def index(app, req):
            return Response('Hello')

        got_request = False,
        got_response = False,

        def middleware(app):
            def handler(env, start_response):
                nonlocal got_request
                got_request = True

                def handle_start_response(*args, **kwargs):
                    nonlocal got_response
                    got_response = True
                    return start_response(*args, **kwargs)

                return app(env, handle_start_response)
            return handler

        builder.add_middleware(middleware)

        app = builder()
        client = Client(app, BaseResponse)

        client.get('/')

        self.assertTrue(got_request)
        self.assertTrue(got_response)

    def test_exception_content_type(self):
        builder = ApplicationBuilder()

        @builder.exception_handler(HTTPException)
        def default_handler(app, req, exc_type, exc_value, exc_traceback):
            return Response('default handler', status=exc_value.code)

        @builder.exception_handler(
            HTTPException, content_type='application/json'
        )
        def default_json_handler(app, req, exc_type, exc_value, exc_traceback):
            return Response(
                '{"type": "json"}',
                status=exc_value.code,
                content_type='application/json'
            )

        @builder.exception_handler(NotFound, content_type='text/html')
        def html_not_found_handler(
                app, req, exc_type, exc_value, exc_traceback):
            return Response('pretty NotFound', status=exc_value.code)

        @builder.expose(route='/raise_418')
        def raise_418(app, req):
            raise ImATeapot()

        @builder.expose(route='/raise_404')
        def raise_404(app, req):
            raise NotFound()

        app = builder()
        client = Client(app, BaseResponse)

        resp = client.get('/raise_418')
        self.assertEqual(resp.status_code, 418)
        self.assertEqual(resp.get_data(), b'{"type": "json"}')

        resp = client.get(
            'raise_418', headers=[('Accept', 'application/json')]
        )
        self.assertEqual(resp.status_code, 418)
        self.assertEqual(resp.get_data(), b'{"type": "json"}')

        resp = client.get(
            'raise_418', headers=[('Accept', 'application/xml')]
        )
        self.assertEqual(resp.get_data(), b'default handler')

        # 404 error has a pretty html representation but uses default renderer
        # for json
        resp = client.get('/raise_404', headers=[('Accept', 'text/html')])
        self.assertEqual(resp.status_code, 404)
        self.assertEqual(resp.get_data(), b'pretty NotFound')

        resp = client.get(
            'raise_404', headers=[('Accept', 'application/json')]
        )
        self.assertEqual(resp.status_code, 404)
        self.assertEqual(resp.get_data(), b'{"type": "json"}')

    def test_close_request(self):
        closed = 0

        class CheckRequestCloseMixin(object):
            def __init__(self):
                def _increment_closed_count():
                    nonlocal closed
                    closed += 1
                self.call_on_close(_increment_closed_count)
                super(CheckRequestCloseMixin, self).__init__()

        builder = ApplicationBuilder()
        builder.add_request_mixins(CheckRequestCloseMixin)

        builder.add_routes(Route('/', endpoint='index'))

        @expose(builder, 'index')
        def index(app, request):
            return Response('Hello World')

        app = builder()
        client = Client(app, BaseResponse)

        # check that requests are closed after success
        client.get('/')
        self.assertEqual(closed, 1)

        # check that requests are closed after an error
        try:
            client.get('/nonexistant')
        except NotFound:
            pass
        self.assertEqual(closed, 2)



"""
    verktyg.urls
    ~~~~~~~~~~~~

    :copyright:
        (c) 2016 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
from urllib.parse import (
    urlsplit, urlunsplit,
    quote as urlquote, quote_plus as urlquote_plus,
)


_hexdigits = '0123456789ABCDEFabcdef'
_hextobyte = dict(
    ((a + b).encode(), int(a + b, 16))
    for a in _hexdigits for b in _hexdigits
)


def _safe_urlunquote_to_bytes(string, unsafe=''):
    if isinstance(string, str):
        string = string.encode('utf-8')
    if isinstance(unsafe, str):
        unsafe = unsafe.encode('utf-8')
    unsafe = frozenset(bytearray(unsafe))
    bits = iter(string.split(b'%'))
    result = bytearray(next(bits, b''))
    for item in bits:
        try:
            char = _hextobyte[item[:2]]
            if char in unsafe:
                raise KeyError()
            result.append(char)
            result.extend(item[2:])
        except KeyError:
            result.extend(b'%')
            result.extend(item)
    return bytes(result)


def _safe_urlunquote(string, charset='utf-8', errors='replace', unsafe=''):
    """URL decode a single string with a given encoding.  If the charset
    is set to `None` no unicode decoding is performed and raw bytes
    are returned.

    :param s:
        The string to unquote.
    :param charset:
        The charset of the query string.  If set to `None` no unicode decoding
        will take place.
    :param errors:
        The error handling for the charset decoding.
    """
    rv = _safe_urlunquote_to_bytes(string, unsafe)
    if charset is not None:
        rv = rv.decode(charset, errors)
    return rv


def encode_idna(domain):
    # If we're given bytes, make sure they fit into ASCII
    if not isinstance(domain, str):
        domain.decode('ascii')
        return domain

    # Otherwise check if it's already ascii, then return
    try:
        return domain.encode('ascii')
    except UnicodeError:
        pass

    # Otherwise encode each part separately
    parts = domain.split('.')
    for idx, part in enumerate(parts):
        parts[idx] = part.encode('idna')
    return b'.'.join(parts)


def decode_idna(domain):
    # If the input is a string try to encode it to ascii to
    # do the idna decoding.  if that fails because of an
    # unicode error, then we already have a decoded idna domain
    if isinstance(domain, str):
        try:
            domain = domain.encode('ascii')
        except UnicodeError:
            return domain

    # Decode each part separately.  If a part fails, try to
    # decode it with ascii and silently ignore errors.  This makes
    # most sense because the idna codec does not have error handling
    parts = domain.split(b'.')
    for idx, part in enumerate(parts):
        try:
            parts[idx] = part.decode('idna')
        except UnicodeError:
            parts[idx] = part.decode('ascii', 'ignore')

    return '.'.join(parts)


def uri_to_iri(uri, errors='replace'):
    r"""Converts a URI in a given charset to a IRI.

    Examples for URI versus IRI:

    >>> uri_to_iri(b'http://xn--n3h.net/')
    u'http://\u2603.net/'
    >>> uri_to_iri(b'http://%C3%BCser:p%C3%A4ssword@xn--n3h.net/p%C3%A5th')
    u'http://\xfcser:p\xe4ssword@\u2603.net/p\xe5th'

    Query strings are left unchanged:

    >>> uri_to_iri('/?foo=24&x=%26%2f')
    u'/?foo=24&x=%26%2f'

    :param uri:
        The URI to convert.
    :param charset:
        The charset of the URI.
    :param errors:
        The error handling on decode.
    """
    assert isinstance(uri, str)
    uri = urlsplit(uri)

    host = decode_idna(uri.hostname) if uri.hostname else ''
    if ':' in host:
        host = '[%s]' % host

    netloc = host

    if uri.port:
        if not 0 <= int(uri.port) <= 65535:
            raise ValueError('Invalid port')
        netloc = '%s:%s' % (netloc, uri.port)

    if uri.username or uri.password:
        if uri.username:
            username = _safe_urlunquote(
                uri.username, errors='strict', unsafe='/:%'
            )
        else:
            username = ''

        if uri.password:
            password = _safe_urlunquote(
                uri.password, errors='strict', unsafe='/:%'
            )
            auth = '%s:%s' % (username, password)
        else:
            auth = username

        netloc = '%s@%s' % (auth, netloc)

    path = _safe_urlunquote(
        uri.path, errors=errors, unsafe='%/;?'
    )
    query = _safe_urlunquote(
        uri.query, errors=errors, unsafe='%;/?:@&=+,$#'
    )
    fragment = _safe_urlunquote(
        uri.fragment, errors=errors, unsafe='%;/?:@&=+,$#'
    )
    return urlunsplit(
        (uri.scheme, netloc, path, query, fragment)
    )


def _encode_netloc(components):
    host = ''
    if components.hostname:
        host = encode_idna(components.hostname).decode('ascii')
    if ':' in host:
        host = '[%s]' % host

    netloc = host

    if components.port:
        if not 0 <= int(components.port) <= 65535:
            raise ValueError('Invalid port')
        netloc = '%s:%s' % (netloc, components.port)

    if components.username or components.password:
        if components.username:
            username = urlquote(
                components.username, safe='/:%'
            )
        else:
            username = ''

        if components.password:
            password = urlquote(
                components.password, safe='/:%'
            )
            auth = '%s:%s' % (username, password)
        else:
            auth = username

        netloc = '%s@%s' % (auth, netloc)
    return netloc


def iri_to_uri(iri):
    r"""Converts any unicode based IRI to an acceptable ASCII URI. Verktyg
    always uses utf-8 URLs internally because this is what browsers and HTTP do
    as well. In some places where it accepts an URL it also accepts a unicode
    IRI and converts it into a URI.

    Examples for IRI versus URI:

    >>> iri_to_uri(u'http://☃.net/')
    'http://xn--n3h.net/'
    >>> iri_to_uri(u'http://üser:pässword@☃.net/påth')
    'http://%C3%BCser:p%C3%A4ssword@xn--n3h.net/p%C3%A5th'

    :param iri:
        The IRI to convert.

    :returns:
        The equivalent URI as an ascii only string object.
    """
    assert isinstance(iri, str)

    iri = urlsplit(iri)

    netloc = _encode_netloc(iri)

    path = urlquote(
        iri.path, safe='/:~+%'
    )
    query = urlquote(
        iri.query, safe='%&[]:;$*()+,!?*/='
    )
    fragment = urlquote(
        iri.fragment, safe='=%&[]:;$()+,!?*/'
    )

    return urlunsplit(
        (iri.scheme, netloc, path, query, fragment)
    )


def url_fix(s, charset='utf-8'):
    r"""Sometimes you get an URL by a user that just isn't a real URL because
    it contains unsafe characters like ' ' and so on. This function can fix
    some of the problems in a similar way browsers handle data entered by the
    user:

    >>> url_fix(u'http://de.wikipedia.org/wiki/Elf (Begriffskl\xe4rung)')
    'http://de.wikipedia.org/wiki/Elf%20(Begriffskl%C3%A4rung)'

    :param s:
        The string with the URL to fix.
    :param charset:
        The target charset for the URL if the url was given as unicode string.
    """
    # First step is to convert backslashes (which are invalid in URLs anyways)
    # to slashes.  This is consistent with what Chrome does.
    s = s.replace('\\', '/')

    # For the specific case that we look like a malformed windows URL
    # we want to fix this up manually:
    if (
        s.startswith('file://') and
        s[7:8].isalpha() and
        s[8:10] in (':/', '|/')
    ):
        s = 'file:///' + s[7:]

    url = urlsplit(s)

    netloc = _encode_netloc(url)

    path = urlquote(
        url.path, encoding=charset, safe='/%+$!*\'(),'
    )
    qs = urlquote_plus(
        url.query, encoding=charset, safe=':&%=+$!*\'(),'
    )
    anchor = urlquote_plus(
        url.fragment, encoding=charset, safe=':&%=+$!*\'(),'
    )

    return urlunsplit(
        (url.scheme, netloc, path, qs, anchor)
    )



"""
    verktyg.http.cache_control
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import re
from hashlib import md5

from verktyg import datastructures
from verktyg.http.basic import (
    is_byte_range_valid, parse_date, dump_header, parse_dict_header, http_date,
)


_etag_re = re.compile(r'([Ww]/)?(?:"(.*?)"|(.*?))(?:\s*,\s*|$)')


class IfRange(object):
    """Very simple object that represents the `If-Range` header in parsed
    form.  It will either have neither a etag or date or one of either but
    never both.
    """

    def __init__(self, etag=None, date=None):
        #: The etag parsed and unquoted.  Ranges always operate on strong
        #: etags so the weakness information is not necessary.
        self.etag = etag
        #: The date in parsed format or `None`.
        self.date = date

    def to_header(self):
        """Converts the object back into an HTTP header."""
        if self.date is not None:
            return http_date(self.date)
        if self.etag is not None:
            return quote_etag(self.etag)
        return ''

    def __str__(self):
        return self.to_header()

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(self))


def parse_if_range_header(value):
    """Parses an if-range header which can be an etag or a date.  Returns
    a :class:`~IfRange` object.
    """
    if not value:
        return IfRange()
    date = parse_date(value)
    if date is not None:
        return IfRange(date=date)
    # drop weakness information
    return IfRange(unquote_etag(value)[0])


class Range(object):
    """Represents a range header.  All the methods are only supporting bytes
    as unit.  It does store multiple ranges but :meth:`range_for_length` will
    only work if only one range is provided.
    """

    def __init__(self, units, ranges):
        #: The units of this range.  Usually "bytes".
        self.units = units
        #: A list of ``(begin, end)`` tuples for the range header provided.
        #: The ranges are non-inclusive.
        self.ranges = ranges

    def range_for_length(self, length):
        """If the range is for bytes, the length is not None and there is
        exactly one range and it is satisfiable it returns a ``(start, stop)``
        tuple, otherwise `None`.
        """
        if self.units != 'bytes' or length is None or len(self.ranges) != 1:
            return None
        start, end = self.ranges[0]
        if end is None:
            end = length
            if start < 0:
                start += length
        if is_byte_range_valid(start, end, length):
            return start, min(end, length)

    def make_content_range(self, length):
        """Creates a :class:`~verktyg.datastructures.ContentRange` object
        from the current range and given content length.
        """
        rng = self.range_for_length(length)
        if rng is not None:
            return ContentRange(self.units, rng[0], rng[1], length)

    def to_header(self):
        """Converts the object back into an HTTP header."""
        ranges = []
        for begin, end in self.ranges:
            if end is None:
                ranges.append(begin >= 0 and '%s-' % begin or str(begin))
            else:
                ranges.append('%s-%s' % (begin, end - 1))
        return '%s=%s' % (self.units, ','.join(ranges))

    def __str__(self):
        return self.to_header()

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(self))


def parse_range_header(value, make_inclusive=True):
    """Parses a range header into a :class:`~Range`
    object.  If the header is missing or malformed `None` is returned.
    `ranges` is a list of ``(start, stop)`` tuples where the ranges are
    non-inclusive.
    """
    if not value or '=' not in value:
        return None

    ranges = []
    last_end = 0
    units, rng = value.split('=', 1)
    units = units.strip().lower()

    for item in rng.split(','):
        item = item.strip()
        if '-' not in item:
            return None
        if item.startswith('-'):
            if last_end < 0:
                return None
            begin = int(item)
            end = None
            last_end = -1
        elif '-' in item:
            begin, end = item.split('-', 1)
            begin = int(begin)
            if begin < last_end or last_end < 0:
                return None
            if end:
                end = int(end) + 1
                if begin >= end:
                    return None
            else:
                end = None
            last_end = end
        ranges.append((begin, end))

    return Range(units, ranges)


class ContentRange(object):
    """Represents the content range header.
    """

    def __init__(self, units, start, stop, length=None, on_update=None):
        if not is_byte_range_valid(start, stop, length):
            raise ValueError('Bad range provided')
        self.on_update = on_update
        self.set(start, stop, length, units)

    def _callback_property(name):
        def fget(self):
            return getattr(self, name)

        def fset(self, value):
            setattr(self, name, value)
            if self.on_update is not None:
                self.on_update(self)
        return property(fget, fset)

    #: The units to use, usually "bytes"
    units = _callback_property('_units')
    #: The start point of the range or `None`.
    start = _callback_property('_start')
    #: The stop point of the range (non-inclusive) or `None`.  Can only be
    #: `None` if also start is `None`.
    stop = _callback_property('_stop')
    #: The length of the range or `None`.
    length = _callback_property('_length')

    def set(self, start, stop, length=None, units='bytes'):
        """Simple method to update the ranges."""
        if not is_byte_range_valid(start, stop, length):
            raise ValueError('Bad range provided')
        self._units = units
        self._start = start
        self._stop = stop
        self._length = length
        if self.on_update is not None:
            self.on_update(self)

    def unset(self):
        """Sets the units to `None` which indicates that the header should
        no longer be used.
        """
        self.set(None, None, units=None)

    def to_header(self):
        if self.units is None:
            return ''
        if self.length is None:
            length = '*'
        else:
            length = self.length
        if self.start is None:
            return '%s */%s' % (self.units, length)
        return '%s %s-%s/%s' % (
            self.units,
            self.start,
            self.stop - 1,
            length
        )

    def __nonzero__(self):
        return self.units is not None

    __bool__ = __nonzero__

    def __str__(self):
        return self.to_header()

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(self))


def parse_content_range_header(value, on_update=None):
    """Parses a range header into a
    :class:`~ContentRange` object or `None` if
    parsing is not possible.

    :param value:
        A content range header to be parsed.
    :param on_update:
        An optional callable that is called every time a value on the
        :class:`~ContentRange` object is changed.
    """
    if value is None:
        return None
    try:
        units, rangedef = (value or '').strip().split(None, 1)
    except ValueError:
        return None

    if '/' not in rangedef:
        return None
    rng, length = rangedef.split('/', 1)
    if length == '*':
        length = None
    elif length.isdigit():
        length = int(length)
    else:
        return None

    if rng == '*':
        return ContentRange(units, None, None, length, on_update=on_update)
    elif '-' not in rng:
        return None

    start, stop = rng.split('-', 1)
    try:
        start = int(start)
        stop = int(stop) + 1
    except ValueError:
        return None

    if is_byte_range_valid(start, stop, length):
        return ContentRange(units, start, stop, length, on_update=on_update)


def cache_property(key, empty, type):
    """Return a new property object for a cache header.  Useful if you
    want to add support for a cache extension in a subclass."""
    return property(
        lambda x: x._get_cache_value(key, empty, type),
        lambda x, v: x._set_cache_value(key, v, type),
        lambda x: x._del_cache_value(key),
        'accessor for %r' % key,
    )


class _CacheControl(datastructures.UpdateDictMixin, dict):
    """Subclass of a dict that stores values for a Cache-Control header.  It
    has accessors for all the cache-control directives specified in RFC 2616.
    The class does not differentiate between request and response directives.

    Because the cache-control directives in the HTTP header use dashes the
    python descriptors use underscores for that.

    To get a header of the :class:`CacheControl` object again you can convert
    the object into a string or call the :meth:`to_header` method.  If you plan
    to subclass it and add your own items have a look at the sourcecode for
    that class.
    """

    no_cache = cache_property('no-cache', '*', None)
    no_store = cache_property('no-store', None, bool)
    max_age = cache_property('max-age', -1, int)
    no_transform = cache_property('no-transform', None, None)

    def __init__(self, values=(), on_update=None):
        dict.__init__(self, values or ())
        self.on_update = on_update
        self.provided = values is not None

    def _get_cache_value(self, key, empty, type):
        """Used internally by the accessor properties."""
        if type is bool:
            return key in self
        if key in self:
            value = self[key]
            if value is None:
                return empty
            elif type is not None:
                try:
                    value = type(value)
                except ValueError:
                    pass
            return value

    def _set_cache_value(self, key, value, type):
        """Used internally by the accessor properties."""
        if type is bool:
            if value:
                self[key] = None
            else:
                self.pop(key, None)
        else:
            if value is None:
                self.pop(key)
            elif value is True:
                self[key] = None
            else:
                self[key] = value

    def _del_cache_value(self, key):
        """Used internally by the accessor properties."""
        if key in self:
            del self[key]

    def to_header(self):
        """Convert the stored values into a cache control header."""
        return dump_header(self)

    def __str__(self):
        return self.to_header()

    def __repr__(self):
        return '<%s %s>' % (
            self.__class__.__name__,
            " ".join(
                "%s=%r" % (k, v) for k, v in sorted(self.items())
            ),
        )


class RequestCacheControl(datastructures.ImmutableDictMixin, _CacheControl):
    """A cache control for requests.  This is immutable and gives access
    to all the request-relevant cache control headers.

    To get a header of the :class:`RequestCacheControl` object again you can
    convert the object into a string or call the :meth:`to_header` method.  If
    you plan to subclass it and add your own items have a look at the
    sourcecode for that class.
    """

    max_stale = cache_property('max-stale', '*', int)
    min_fresh = cache_property('min-fresh', '*', int)
    no_transform = cache_property('no-transform', None, None)
    only_if_cached = cache_property('only-if-cached', None, bool)


class ResponseCacheControl(_CacheControl):
    """A cache control for responses.  Unlike :class:`RequestCacheControl`
    this is mutable and gives access to response-relevant cache control
    headers.

    To get a header of the :class:`ResponseCacheControl` object again you can
    convert the object into a string or call the :meth:`to_header` method.  If
    you plan to subclass it and add your own items have a look at the
    sourcecode for that class.
    """

    public = cache_property('public', None, bool)
    private = cache_property('private', '*', None)
    must_revalidate = cache_property('must-revalidate', None, bool)
    proxy_revalidate = cache_property('proxy-revalidate', None, bool)
    s_maxage = cache_property('s-maxage', None, None)


# attach cache_property to the _CacheControl as staticmethod
# so that others can reuse it.
_CacheControl.cache_property = staticmethod(cache_property)


def parse_cache_control_header(value, on_update=None, cls=None):
    """Parse a cache control header.  The RFC differs between response and
    request cache control, this method does not.  It's your responsibility
    to not use the wrong control statements.

    :param value:
        A cache control header to be parsed.
    :param on_update:
        An optional callable that is called every time a value on the
        :class:`~verktyg.CacheControl` object is changed.
    :param cls:
        The class for the returned object.  By default
        :class:`~verktyg.RequestCacheControl` is used.
    :return:
        A `cls` object.
    """
    if cls is None:
        cls = RequestCacheControl
    if not value:
        return cls(None, on_update)
    return cls(parse_dict_header(value), on_update)


def quote_etag(etag, weak=False):
    """Quote an etag.

    :param etag:
        The etag to quote.
    :param weak:
        Set to `True` to tag it "weak".
    """
    if '"' in etag:
        raise ValueError('invalid etag')
    etag = '"%s"' % etag
    if weak:
        etag = 'w/' + etag
    return etag


def unquote_etag(etag):
    """Unquote a single etag:

    >>> unquote_etag('w/"bar"')
    ('bar', True)
    >>> unquote_etag('"bar"')
    ('bar', False)

    :param etag:
        The etag identifier to unquote.
    :return:
        An ``(etag, weak)`` tuple.
    """
    if not etag:
        return None, None
    etag = etag.strip()
    weak = False
    if etag[:2] in ('w/', 'W/'):
        weak = True
        etag = etag[2:]
    if etag[:1] == etag[-1:] == '"':
        etag = etag[1:-1]
    return etag, weak


class ETags(object):
    """A set that can be used to check if one etag is present in a collection
    of etags.
    """

    def __init__(self, strong_etags=None, weak_etags=None, star_tag=False):
        self._strong = frozenset(not star_tag and strong_etags or ())
        self._weak = frozenset(weak_etags or ())
        self.star_tag = star_tag

    def as_set(self, include_weak=False):
        """Convert the `ETags` object into a python set.  Per default all the
        weak etags are not part of this set."""
        rv = set(self._strong)
        if include_weak:
            rv.update(self._weak)
        return rv

    def is_weak(self, etag):
        """Check if an etag is weak."""
        return etag in self._weak

    def contains_weak(self, etag):
        """Check if an etag is part of the set including weak and strong tags.
        """
        return self.is_weak(etag) or self.contains(etag)

    def contains(self, etag):
        """Check if an etag is part of the set ignoring weak tags.
        It is also possible to use the ``in`` operator.

        """
        if self.star_tag:
            return True
        return etag in self._strong

    def contains_raw(self, etag):
        """When passed a quoted tag it will check if this tag is part of the
        set.  If the tag is weak it is checked against weak and strong tags,
        otherwise strong only."""
        etag, weak = unquote_etag(etag)
        if weak:
            return self.contains_weak(etag)
        return self.contains(etag)

    def to_header(self):
        """Convert the etags set into a HTTP header string."""
        if self.star_tag:
            return '*'
        return ', '.join(
            ['"%s"' % x for x in self._strong] +
            ['w/"%s"' % x for x in self._weak]
        )

    def __call__(self, etag=None, data=None, include_weak=False):
        if [etag, data].count(None) != 1:
            raise TypeError('either tag or data required, but at least one')
        if etag is None:
            etag = generate_etag(data)
        if include_weak:
            if etag in self._weak:
                return True
        return etag in self._strong

    def __bool__(self):
        return bool(self.star_tag or self._strong or self._weak)

    __nonzero__ = __bool__

    def __str__(self):
        return self.to_header()

    def __iter__(self):
        return iter(self._strong)

    def __contains__(self, etag):
        return self.contains(etag)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(self))


def parse_etags(value):
    """Parse an etag header.

    :param value:
        The tag header to parse
    :return:
        An :class:`~ETags` object.
    """
    if not value:
        return ETags()
    strong = []
    weak = []
    end = len(value)
    pos = 0
    while pos < end:
        match = _etag_re.match(value, pos)
        if match is None:
            break
        is_weak, quoted, raw = match.groups()
        if raw == '*':
            return ETags(star_tag=True)
        elif quoted:
            raw = quoted
        if is_weak:
            weak.append(raw)
        else:
            strong.append(raw)
        pos = match.end()
    return ETags(strong, weak)


def generate_etag(data):
    """Generate an etag for some data."""
    return md5(data).hexdigest()


def is_resource_modified(environ, etag=None, data=None, last_modified=None):
    """Convenience method for conditional requests.

    :param environ:
        The WSGI environment of the request to be checked.
    :param etag:
        The etag for the response for comparison.
    :param data:
        Or alternatively the data of the response to automatically generate an
        etag using :func:`generate_etag`.
    :param last_modified:
        An optional date of the last modification.
    :return:
        `True` if the resource was modified, otherwise `False`.
    """
    if etag is None and data is not None:
        etag = generate_etag(data)
    elif data is not None:
        raise TypeError('both data and etag given')
    if environ['REQUEST_METHOD'] not in ('GET', 'HEAD'):
        return False

    unmodified = False
    if isinstance(last_modified, str):
        last_modified = parse_date(last_modified)

    # ensure that microsecond is zero because the HTTP spec does not transmit
    # that either and we might have some false positives.  See issue #39
    if last_modified is not None:
        last_modified = last_modified.replace(microsecond=0)

    modified_since = parse_date(environ.get('HTTP_IF_MODIFIED_SINCE'))

    if modified_since and last_modified and last_modified <= modified_since:
        unmodified = True
    if etag:
        if_none_match = parse_etags(environ.get('HTTP_IF_NONE_MATCH'))
        if if_none_match:
            unmodified = if_none_match.contains_raw(etag)

    return not unmodified



"""
    verktyg.request
    ~~~~~~~~~~~~~~~

    The wrappers are simple request and response objects which you can
    subclass to do whatever you want them to do.  The request object contains
    the information transmitted by the client (webbrowser) and the response
    object contains all the information sent back to the browser.

    An important detail is that the request object is created with the WSGI
    environ and will act as high-level proxy whereas the response object is an
    actual WSGI application.

    Like everything else in Werkzeug these objects will work correctly with
    unicode data.  Incoming form data parsed by the response object will be
    decoded into an unicode object if possible and if it makes sense.


    :copyright: (c) 2014 by the Werkzeug Team, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""
from io import BytesIO
from urllib.parse import parse_qsl

from verktyg.utils import cached_property
from verktyg.datastructures import (
    ImmutableMultiDict,
    ImmutableTypeConversionDict,
    ImmutableList,
    iter_multi_items,
)
from verktyg.http import (
    parse_cache_control_header, parse_etags,
    parse_date, parse_set_header, parse_authorization_header,
    parse_options_header, parse_if_range_header, parse_cookie,
    parse_range_header,
    RequestCacheControl,
)
from verktyg.wsgi import (
    EnvironHeaders, wsgi_decoding_dance,
    get_current_url, get_host, get_input_stream, get_content_length
)


def _assert_not_shallow(request):
    if request.shallow:
        raise RuntimeError(
            'A shallow request tried to consume form data.  If you really '
            'want to do that, set `shallow` to False.'
        )


class BaseRequest(object):

    """Very basic request object.  This does not implement advanced stuff like
    entity tag parsing or cache controls.  The request object is created with
    the WSGI environment as first argument and will add itself to the WSGI
    environment as ``'verktyg.request'`` unless it's created with
    `populate_request` set to False.

    There are a couple of mixins available that add additional functionality
    to the request object, there is also a class called `Request` which
    subclasses `BaseRequest` and all the important mixins.

    It's a good idea to create a custom subclass of the :class:`BaseRequest`
    and add missing functionality either via mixins or direct implementation.
    Here an example for such subclasses::

        from verktyg.wrappers import BaseRequest, ETagRequestMixin

        class Request(BaseRequest, ETagRequestMixin):
            pass

    Request objects are **read only**.  As of 0.5 modifications are not
    allowed in any place.  Unlike the lower level parsing functions the
    request object will use immutable objects everywhere possible.

    Per default the request object will assume all the text data is `utf-8`
    encoded.  Please refer to `the unicode chapter <unicode.txt>`_ for more
    details about customizing the behavior.

    Per default the request object will be added to the WSGI
    environment as `verktyg.request` to support the debugging system.
    If you don't want that, set `populate_request` to `False`.

    If `shallow` is `True` the environment is initialized as shallow
    object around the environ.  Every operation that would modify the
    environ in any way (such as consuming form data) raises an exception
    unless the `shallow` attribute is explicitly set to `False`.  This
    is useful for middlewares where you don't want to consume the form
    data by accident.  A shallow request is not populated to the WSGI
    environment.
    """
    #: the error handling procedure for errors, defaults to 'replace'
    encoding_errors = 'replace'

    #: the maximum content length.  This is forwarded to the form data
    #: parsing function (:func:`parse_form_data`).  When set and the
    #: :attr:`form` or :attr:`files` attribute is accessed and the
    #: parsing fails because more than the specified value is transmitted
    #: a :exc:`~verktyg.exceptions.RequestEntityTooLarge` exception is raised.
    #:
    #: Have a look at :ref:`dealing-with-request-data` for more details.
    max_content_length = None

    #: the maximum form field size.  This is forwarded to the form data
    #: parsing function (:func:`parse_form_data`).  When set and the
    #: :attr:`form` or :attr:`files` attribute is accessed and the
    #: data in memory for post data is longer than the specified value a
    #: :exc:`~verktyg.exceptions.RequestEntityTooLarge` exception is raised.
    #:
    #: Have a look at :ref:`dealing-with-request-data` for more details.
    max_form_memory_size = None

    #: the class to use for `args` and `form`.  The default is an
    #: :class:`~verktyg.datastructures.ImmutableMultiDict` which supports
    #: multiple values per key.  alternatively it makes sense to use an
    #: :class:`~verktyg.datastructures.ImmutableOrderedMultiDict` which
    #: preserves order or a :class:`~verktyg.datastructures.ImmutableDict`
    #: which is the fastest but only remembers the last key.  It is also
    #: possible to use mutable structures, but this is not recommended.
    parameter_storage_class = ImmutableMultiDict

    #: the type to be used for list values from the incoming WSGI environment.
    #: By default an :class:`~verktyg.datastructures.ImmutableList` is used
    #: (for example for :attr:`access_list`).
    list_storage_class = ImmutableList

    #: the type to be used for dict values from the incoming WSGI environment.
    #: By default an
    #: :class:`~verktyg.datastructures.ImmutableTypeConversionDict` is used
    #: (for example for :attr:`cookies`).
    dict_storage_class = ImmutableTypeConversionDict

    #: Optionally a list of hosts that is trusted by this request.  By default
    #: all hosts are trusted which means that whatever the client sends the
    #: host is will be accepted.
    #:
    #: This is the recommended setup as a webserver should manually be set up
    #: to only route correct hosts to the application, and remove the
    #: `X-Forwarded-Host` header if it is not being used (see
    #: :func:`verktyg.wsgi.get_host`).
    trusted_hosts = None

    #: Indicates whether the data descriptor should be allowed to read and
    #: buffer up the input stream.  By default it's enabled.
    disable_data_descriptor = False

    def __init__(self, environ, populate_request=True, shallow=False):
        self.environ = environ
        if populate_request and not shallow:
            self.environ['verktyg.request'] = self
        self.shallow = shallow
        self._on_close = []
        super(BaseRequest, self).__init__()

    def __repr__(self):
        # make sure the __repr__ even works if the request was created
        # from an invalid WSGI environment.  If we display the request
        # in a debug session we don't want the repr to blow up.
        args = []
        try:
            args.append("'%s'" % self.url)
            args.append('[%s]' % self.method)
        except Exception:
            args.append('(invalid WSGI environ)')

        return '<%s %s>' % (
            self.__class__.__name__,
            ' '.join(args)
        )

    @property
    def app(self):
        """The verktyg application the created the request

        :return: a `verktyg.application.Application` object or `None`
        """
        return self.environ.get('verktyg.application')

    @classmethod
    def from_values(cls, *args, **kwargs):
        """Create a new request object based on the values provided.  If
        environ is given missing values are filled from there.  This method is
        useful for small scripts when you need to simulate a request from an
        URL.
        Do not use this method for unittesting, there is a full featured client
        object (:class:`Client`) that allows to create multipart requests,
        support for cookies etc.

        This accepts the same options as the
        :class:`~verktyg.test.EnvironBuilder`.
        :return: request object
        """
        from verktyg.test import EnvironBuilder
        builder = EnvironBuilder(*args, **kwargs)
        try:
            return builder.get_request(cls)
        finally:
            builder.close()

    def _get_stream_for_parsing(self):
        """This is the same as accessing :attr:`stream` with the difference
        that if it finds cached data from calling :meth:`get_data` first it
        will create a new stream out of the cached data.
        """
        cached_data = getattr(self, '_cached_data', None)
        if cached_data is not None:
            return BytesIO(cached_data)
        return self.stream

    def call_on_close(self, func):
        """Adds a function to the internal list of functions that should
        be called as part of closing down the request.  Also returns the
        function that was passed so that this can be used as a decorator.
        """
        self._on_close.append(func)
        return func

    def close(self):
        """Closes associated resources of this request object.  This
        closes all file handles explicitly.  You can also use the request
        object in a with statement which will automatically close it.
        """
        files = self.__dict__.get('files')
        for key, value in iter_multi_items(files or ()):
            value.close()
        for func in self._on_close:
            func()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    @cached_property
    def stream(self):
        """The stream to read incoming data from.  Unlike :attr:`input_stream`
        this stream is properly guarded that you can't accidentally read past
        the length of the input.  Werkzeug will internally always refer to
        this stream to read data which makes it possible to wrap this
        object with a stream that does filtering.
        """
        _assert_not_shallow(self)
        return get_input_stream(self.environ)

    @cached_property
    def input_stream(self):
        """The WSGI input stream.

        In general it's a bad idea to use this one because you can easily read
        past the boundary.  Use the :attr:`stream` instead.
        """
        return self.environ.get('wsgi.input')

    @cached_property
    def args(self):
        """The parsed URL parameters.  By default an
        :class:`~verktyg.datastructures.ImmutableMultiDict`
        is returned from this function.  This can be changed by setting
        :attr:`parameter_storage_class` to a different type.  This might
        be necessary if the order of the form data is important.
        """
        qs = wsgi_decoding_dance(
            self.environ.get('QUERY_STRING', ''),
            errors=self.encoding_errors,
        )
        return self.parameter_storage_class(parse_qsl(
            qs, errors=self.encoding_errors,
        ))

    def get_data(self, cache=True, as_text=False):
        """This reads the buffered incoming data from the client into one
        bytestring.  By default this is cached but that behavior can be
        changed by setting `cache` to `False`.

        Usually it's a bad idea to call this method without checking the
        content length first as a client could send dozens of megabytes or more
        to cause memory problems on the server.

        If `as_text` is set to `True` the return value will be a decoded
        unicode string.
        """
        rv = getattr(self, '_cached_data', None)
        if rv is None:
            rv = self.stream.read()
            if cache:
                self._cached_data = rv
        if as_text:
            # TODO charset
            rv = rv.decode('utf-8', self.encoding_errors)
        return rv

    @cached_property
    def cookies(self):
        """Read only access to the retrieved cookie values as dictionary."""
        return parse_cookie(
            self.environ, errors=self.encoding_errors,
            cls=self.dict_storage_class
        )

    @cached_property
    def headers(self):
        """The headers from the WSGI environ as immutable
        :class:`~verktyg.wsgi.EnvironHeaders`.
        """
        return EnvironHeaders(self.environ)

    @cached_property
    def path(self):
        """Requested path as unicode.  This works a bit like the regular path
        info in the WSGI environment but will always include a leading slash,
        even if the URL root is accessed.
        """
        raw_path = wsgi_decoding_dance(
            self.environ.get('PATH_INFO') or '', errors=self.encoding_errors
        )
        return '/' + raw_path.lstrip('/')

    @cached_property
    def full_path(self):
        """Requested path as unicode, including the query string."""
        return self.path + u'?' + self.query_string

    @cached_property
    def script_root(self):
        """The root path of the script without the trailing slash."""
        raw_path = wsgi_decoding_dance(
            self.environ.get('SCRIPT_NAME') or '',
            errors=self.encoding_errors
        )
        return raw_path.rstrip('/')

    @cached_property
    def url(self):
        """The reconstructed current URL as IRI.
        See also: :attr:`trusted_hosts`.
        """
        return get_current_url(
            self.environ, trusted_hosts=self.trusted_hosts,
        )

    @cached_property
    def base_url(self):
        """Like :attr:`url` but without the querystring
        See also: :attr:`trusted_hosts`.
        """
        return get_current_url(
            self.environ, strip_querystring=True,
            trusted_hosts=self.trusted_hosts,
        )

    @cached_property
    def url_root(self):
        """The full URL root (with hostname), this is the application
        root as IRI.
        See also: :attr:`trusted_hosts`.
        """
        return get_current_url(
            self.environ, True, trusted_hosts=self.trusted_hosts,
        )

    @cached_property
    def host_url(self):
        """Just the host with scheme as IRI.
        See also: :attr:`trusted_hosts`.
        """
        return get_current_url(
            self.environ, host_only=True, trusted_hosts=self.trusted_hosts,
        )

    @cached_property
    def host(self):
        """Just the host including the port if available.
        See also: :attr:`trusted_hosts`.
        """
        return get_host(self.environ, trusted_hosts=self.trusted_hosts)

    @cached_property
    def query_string(self):
        return wsgi_decoding_dance(
            self.environ.get('QUERY_STRING') or '', errors=self.encoding_errors
        )

    @cached_property
    def method(self):
        """The transmission method. (For example ``'GET'`` or ``'POST'``).
        """
        return self.environ.get('REQUEST_METHOD', 'GET').upper()

    @cached_property
    def access_route(self):
        """If a forwarded header exists this is a list of all ip addresses
        from the client ip to the last proxy server.
        """
        if 'HTTP_X_FORWARDED_FOR' in self.environ:
            addr = self.environ['HTTP_X_FORWARDED_FOR'].split(',')
            return self.list_storage_class([x.strip() for x in addr])
        elif 'REMOTE_ADDR' in self.environ:
            return self.list_storage_class([self.environ['REMOTE_ADDR']])
        return self.list_storage_class()

    @property
    def remote_addr(self):
        """The remote address of the client."""
        return self.environ.get('REMOTE_ADDR')

    @cached_property
    def remote_user(self):
        """If the server supports user authentication, and the script is
        protected, this attribute contains the username the user has
        authenticated as.
        """
        return self.environ.get('REMOTE_USER')

    @cached_property
    def scheme(self):
        """URL scheme (http or https).
        """
        return self.environ.get('wsgi.url_scheme')

    @property
    def is_xhr(self):
        """True if the request was triggered via a JavaScript XMLHttpRequest.

        This only works with libraries that support the `X-Requested-With`
        header and set it to "XMLHttpRequest".  Libraries that do that are
        prototype, jQuery and Mochikit and probably some more.
        """
        requested_with = self.environ.get('HTTP_X_REQUESTED_WITH', '').lower()
        return requested_with.lower() == 'xmlhttprequest'

    @property
    def is_secure(self):
        """`True` if the request is made over an encrypted connection
        """
        self.environ['wsgi.url_scheme'] == 'https',

    @property
    def is_multithread(self):
        """`True` if the application is served by a multithreaded WSGI server.
        """
        return self.environ.get('wsgi.multithread')

    @property
    def is_multiprocess(self):
        """`True` if the application is served by a multiprocess WSGI server.
        """
        return self.environ.get('wsgi.multiprocess')

    @property
    def is_run_once(self):
        """`True` if the application will be executed only once in a process
        lifetime.

        This is the case for CGI for example, but it's not guaranteed that the
        execution only happens one time.
        """
        return self.environ.get('wsgi.run_once')


class ETagRequestMixin(object):

    """Add entity tag and cache descriptors to a request object or object with
    a WSGI environment available as :attr:`~BaseRequest.environ`.  This not
    only provides access to etags but also to the cache control header.
    """

    @cached_property
    def cache_control(self):
        """A :class:`~verktyg.http.RequestCacheControl` object
        for the incoming cache control headers.
        """
        cache_control = self.environ.get('HTTP_CACHE_CONTROL')
        return parse_cache_control_header(
            cache_control, None, RequestCacheControl,
        )

    @cached_property
    def if_match(self):
        """An object containing all the etags in the `If-Match` header.

        :rtype: :class:`~verktyg.http.ETags`
        """
        return parse_etags(self.environ.get('HTTP_IF_MATCH'))

    @cached_property
    def if_none_match(self):
        """An object containing all the etags in the `If-None-Match` header.

        :rtype: :class:`~verktyg.http.ETags`
        """
        return parse_etags(self.environ.get('HTTP_IF_NONE_MATCH'))

    @cached_property
    def if_modified_since(self):
        """The parsed `If-Modified-Since` header as datetime object."""
        return parse_date(self.environ.get('HTTP_IF_MODIFIED_SINCE'))

    @cached_property
    def if_unmodified_since(self):
        """The parsed `If-Unmodified-Since` header as datetime object."""
        return parse_date(self.environ.get('HTTP_IF_UNMODIFIED_SINCE'))

    @cached_property
    def if_range(self):
        """The parsed `If-Range` header.

        :rtype: :class:`~verktyg.http.IfRange`
        """
        return parse_if_range_header(self.environ.get('HTTP_IF_RANGE'))

    @cached_property
    def range(self):
        """The parsed `Range` header.

        :rtype: :class:`~verktyg.http.Range`
        """
        return parse_range_header(self.environ.get('HTTP_RANGE'))


class AuthorizationMixin(object):

    """Adds an :attr:`authorization` property that represents the parsed
    value of the `Authorization` header as
    :class:`~verktyg.http.Authorization` object.
    """

    @cached_property
    def authorization(self):
        """The `Authorization` object in parsed form."""
        header = self.environ.get('HTTP_AUTHORIZATION')
        return parse_authorization_header(header)


class StreamOnlyMixin(object):

    """If mixed in before the request object this will change the behavior
    of it to disable handling of form parsing.  This disables the
    :attr:`files`, :attr:`form` attributes and will just provide a
    :attr:`stream` attribute that is always available.
    """

    disable_data_descriptor = True
    want_form_data_parsed = False


class CommonRequestDescriptorsMixin(object):

    """A mixin for :class:`BaseRequest` subclasses.  Request objects that
    mix this class in will automatically get descriptors for a couple of
    HTTP headers with automatic type conversion.
    """

    @property
    def content_type(self):
        """The Content-Type entity-header field indicates the media type of
        the entity-body sent to the recipient or, in the case of the HEAD
        method, the media type that would have been sent had the request
        been a GET.
        """
        return self.environ.get('CONTENT_TYPE')

    @cached_property
    def content_length(self):
        """The Content-Length entity-header field indicates the size of the
        entity-body in bytes or, in the case of the HEAD method, the size of
        the entity-body that would have been sent had the request been a
        GET.
        """
        return get_content_length(self.environ)

    @property
    def content_encoding(self):
        """The Content-Encoding entity-header field is used as a modifier to
        the media-type.  When present, its value indicates what additional
        content codings have been applied to the entity-body, and thus what
        decoding mechanisms must be applied in order to obtain the media-type
        referenced by the Content-Type header field.
        """
        return self.environ.get('HTTP_CONTENT_ENCODING')

    @property
    def content_md5(self):
        """The Content-MD5 entity-header field, as defined in RFC 1864, is an
        MD5 digest of the entity-body for the purpose of providing an
        end-to-end message integrity check (MIC) of the entity-body.  (Note:
        a MIC is good for detecting accidental modification of the
        entity-body in transit, but is not proof against malicious attacks.)
        """
        return self.environ.get('HTTP_CONTENT_MD5')

    @property
    def referrer(self):
        """The Referer[sic] request-header field allows the client to specify,
        for the server's benefit, the address (URI) of the resource from which
        the Request-URI was obtained (the "referrer", although the header
        field is misspelled).
        """
        return self.environ.get('HTTP_REFERER')

    @property
    def date(self):
        """The Date general-header field represents the date and time at which
        the message was originated, having the same semantics as orig-date
        in RFC 822.
        """
        return parse_date(self.environ.get('HTTP_DATE'))

    @property
    def max_forwards(self):
        """The Max-Forwards request-header field provides a mechanism with the
        TRACE and OPTIONS methods to limit the number of proxies or gateways
        that can forward the request to the next inbound server.
        """
        try:
            return int(self.environ.get('HTTP_MAX_FORWARDS'))
        except ValueError:
            return None

    def _parse_content_type(self):
        if not hasattr(self, '_parsed_content_type'):
            self._parsed_content_type = parse_options_header(
                self.environ.get('CONTENT_TYPE', '')
            )

    @property
    def mimetype(self):
        """Like :attr:`content_type`, but without parameters (eg, without
        charset, type etc.) and always lowercase.  For example if the content
        type is ``text/HTML; charset=utf-8`` the mimetype would be
        ``'text/html'``.
        """
        self._parse_content_type()
        return self._parsed_content_type[0].lower()

    @property
    def mimetype_params(self):
        """The mimetype parameters as dict.  For example if the content
        type is ``text/html; charset=utf-8`` the params would be
        ``{'charset': 'utf-8'}``.
        """
        self._parse_content_type()
        return self._parsed_content_type[1]

    @cached_property
    def pragma(self):
        """The Pragma general-header field is used to include
        implementation-specific directives that might apply to any recipient
        along the request/response chain.  All pragma directives specify
        optional behavior from the viewpoint of the protocol; however, some
        systems MAY require that behavior be consistent with the directives.
        """
        return parse_set_header(self.environ.get('HTTP_PRAGMA', ''))


# TODO deprecate.  Superseded by request class building in ApplicationBuilder
class Request(
    BaseRequest,
    ETagRequestMixin,
    AuthorizationMixin,
    CommonRequestDescriptorsMixin,
):

    """Full featured request object implementing the following mixins:

    - :class:`AcceptMixin` for accept header parsing
    - :class:`ETagRequestMixin` for etag and cache control handling
    - :class:`AuthorizationMixin` for http auth handling
    - :class:`CommonRequestDescriptorsMixin` for common headers
    """


class PlainRequest(StreamOnlyMixin, Request):

    """A request object without special form parsing capabilities.
    """



"""
    verktyg.testsuite.test_accept_content_type
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests for HTTP accept header parsing utilities.


    :copyright:
        (c) 2016 Ben Mather
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

from verktyg.accept.content_type import (
    parse_content_type_header, parse_accept_header,
    ContentType, ContentTypeAccept,
)
from verktyg.exceptions import NotAcceptable


class ContentTypeTestCase(unittest.TestCase):
    def test_parse_accept_basic(self):
        accept = parse_accept_header(
            'text/xml'
        )

        range_ = next(iter(accept))
        self.assertEqual('text', range_.type)
        self.assertEqual('xml', range_.subtype)
        self.assertEqual(1, range_.q)
        self.assertEqual(0, len(range_.params))
        with self.assertRaises(KeyError):
            range_.params['q']

    def test_parse_accept_params(self):
        accept = parse_accept_header(
            'application/foo;quiet=no; bar=baz;q=0.6'
        )
        range_ = next(iter(accept))
        self.assertEqual('application', range_.type)
        self.assertEqual('foo', range_.subtype)
        self.assertEqual(0.6, range_.q)

        self.assertEqual(2, len(range_.params))
        self.assertEqual('no', range_.params['quiet'])
        self.assertEqual('baz', range_.params['bar'])
        with self.assertRaises(KeyError):
            range_.params['no-such-param']
        with self.assertRaises(KeyError):
            range_.params['q']

    def test_parse_accept_invalid_params(self):
        # TODO
        pass

    def test_parse_accept_case(self):
        # TODO
        pass

    def test_parse_accept_multiple(self):
        accept = parse_accept_header(
            'text/xml,'
            'application/xml,'
            'application/xhtml+xml,'
            'application/foo;quiet=no; bar=baz;q=0.6,'
            'text/html;q=0.9,'
            'text/plain;q=0.8,'
            'image/png,'
            '*/*;q=0.5'
        )

        self.assertEqual(len(list(accept)), 8)

    def test_parse(self):
        content_type = parse_content_type_header('text/html')

        self.assertEqual(content_type.type, 'text')
        self.assertEqual(content_type.subtype, 'html')

    def test_serialize(self):
        content_type = ContentType('text/html', qs=0.5)

        self.assertEqual('text/html', content_type.to_header())

    def test_serialize_accept(self):
        accept = ContentTypeAccept(['text/html'])

        self.assertEqual(accept.to_header(), 'text/html')

    def test_serialize_accept_q_before_params(self):
        accept = ContentTypeAccept([
            ('application/json', '0.5', {'speed': 'maximum'}),
        ])

        self.assertEqual(
            accept.to_header(), 'application/json;q=0.5;speed=maximum'
        )

    def test_serialize_accept_redundant_q(self):
        accept = ContentTypeAccept([('image/png', '1')])
        self.assertEqual(accept.to_header(), 'image/png')

    def test_serialize_accept_multiple(self):
        accept = ContentTypeAccept([
            'application/xhtml+xml',
            ('text/plain', '0.8'),
            'image/png',
            ('*/*', '0.5'),
        ])
        self.assertEqual(
            accept.to_header(),
            (
                'application/xhtml+xml,'
                'text/plain;q=0.8,'
                'image/png,'
                '*/*;q=0.5'
            )
        )

    def test_match_basic(self):
        accept = ContentTypeAccept(['text/xml'])

        acceptable = ContentType('text/xml')
        unacceptable_type = ContentType('application/xml')
        unacceptable_subtype = ContentType('text/html')

        self.assertRaises(
            NotAcceptable, unacceptable_type.acceptability, accept
        )
        self.assertRaises(
            NotAcceptable, unacceptable_subtype.acceptability, accept
        )

        match = acceptable.acceptability(accept)
        self.assertEqual(acceptable, match.content_type)
        self.assertTrue(match.type_matches)
        self.assertTrue(match.subtype_matches)
        self.assertTrue(match.exact_match)

    def test_match_wildcard(self):
        accept = ContentTypeAccept(['*/*'])

        content_type = ContentType('text/html')

        match = content_type.acceptability(accept)
        self.assertEqual(content_type, match.content_type)
        self.assertFalse(match.type_matches)
        self.assertFalse(match.subtype_matches)
        self.assertFalse(match.exact_match)

    def test_match_subtype_wildcard(self):
        accept = ContentTypeAccept(['text/*'])

        unacceptable = ContentType('image/jpeg')
        acceptable = ContentType('text/html')

        self.assertRaises(NotAcceptable, unacceptable.acceptability, accept)

        match = acceptable.acceptability(accept)
        self.assertEqual(acceptable, match.content_type)
        self.assertTrue(match.type_matches)
        self.assertFalse(match.subtype_matches)
        self.assertFalse(match.exact_match)

    def test_match_quality(self):
        accept = ContentTypeAccept([('text/html', '0.5')])

        no_qs = ContentType('text/html')
        qs = ContentType('text/html', qs=0.5)

        self.assertEqual(0.5, no_qs.acceptability(accept).quality)
        self.assertEqual(0.25, qs.acceptability(accept).quality)

    def test_match(self):
        accept = ContentTypeAccept([
            'text/xml',
            'application/xml',
            'application/xhtml+xml',
            ('application/foo', '0.6', {'quiet': 'no', 'bar': 'baz'}),
            ('text/html', '0.9'),
            ('text/plain', '0.8'),
            'image/png',
            ('*/*', '0.5'),
        ])

        content_type = ContentType('image/png')
        match = content_type.acceptability(accept)
        self.assertEqual(match.quality, 1.0)
        self.assertTrue(match.exact_match)

        content_type = ContentType('text/plain')
        match = content_type.acceptability(accept)
        self.assertEqual(match.quality, 0.8)
        self.assertTrue(match.exact_match)

        content_type = ContentType('application/json')
        match = content_type.acceptability(accept)
        self.assertEqual(match.quality, 0.5)
        self.assertFalse(match.exact_match)



"""
    verktyg.testsuite.test_accept_language
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests for HTTP accept header parsing utilities.


    :copyright:
        (c) 2016 Ben Mather
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

from verktyg.accept.language import (
    parse_language_header, parse_accept_language_header,
    Language, LanguageAccept,
)

from verktyg.exceptions import NotAcceptable


class LanguageTestCase(unittest.TestCase):
    def test_parse_accept_basic(self):
        accept = parse_accept_language_header('en-gb')

        range_ = next(iter(accept))
        self.assertEqual('en-gb', range_.value)
        self.assertEqual(1, range_.q)

    def test_parse_accept_params(self):
        self.assertRaises(
            ValueError, parse_accept_language_header, 'en;param=invalid'
        )

    def test_parse_accept_q(self):
        accept = parse_accept_language_header('en; q=0.8')
        range_ = next(iter(accept))

        self.assertEqual(0.8, range_.q)

    def test_parse_accept_multiple(self):
        accept = parse_accept_language_header(
            'da,en-gb;q=0.8,en;q=0.7,*;q=0.1'
        )

        self.assertEqual(len(list(accept)), 4)

    def test_parse(self):
        language = parse_language_header('en-gb')

        self.assertEqual(language.value, 'en-gb')

    def test_serialize(self):
        language = Language('en-us', qs=0.5)

        self.assertEqual('en-us', language.to_header())

    def test_serialize_accept_redundant_q(self):
        accept = LanguageAccept([('jp', '1')])
        self.assertEqual(accept.to_header(), 'jp')

    def test_serialize_accept_multiple(self):
        accept = LanguageAccept([
            'da', ('en-gb', 0.8), ('en', 0.7), ('*', 0.1)
        ])

        self.assertEqual(accept.to_header(), 'da,en-gb;q=0.8,en;q=0.7,*;q=0.1')

    def test_match_basic(self):
        accept = LanguageAccept(['en-gb'])

        acceptable = Language('en-gb')
        unacceptable = Language('fr')

        self.assertRaises(NotAcceptable, unacceptable.acceptability, accept)

        match = acceptable.acceptability(accept)
        self.assertEqual(match.language, acceptable)
        self.assertEqual(match.specificity, 2)
        self.assertEqual(match.tail, 0)
        self.assertTrue(match.exact_match)

    def test_match_partial(self):
        accept = LanguageAccept(['one-two'])

        unacceptable = Language('one')
        acceptable = Language('one-two-three')

        self.assertRaises(NotAcceptable, unacceptable.acceptability, accept)

        match = acceptable.acceptability(accept)
        self.assertEqual(acceptable, match.language)
        self.assertEqual(match.specificity, 2)
        self.assertEqual(match.tail, 1)
        self.assertFalse(match.exact_match)

    def test_match_wildcard(self):
        accept = LanguageAccept(['*'])

        language = Language('en')

        match = language.acceptability(accept)
        self.assertEqual(match.language, language)
        self.assertEqual(match.specificity, 0)
        self.assertEqual(match.tail, 1)
        self.assertFalse(match.exact_match)

    def test_match_quality(self):
        accept = LanguageAccept([('en', '0.5')])

        no_qs = Language('en')
        qs = Language('en', qs=0.5)

        self.assertEqual(0.5, no_qs.acceptability(accept).quality)
        self.assertEqual(0.25, qs.acceptability(accept).quality)

    def test_match(self):
        accept = LanguageAccept([
            'fr', 'fr-be', ('en-gb', 0.8), ('en', 0.7), ('*', 0.1)
        ])

        fr = Language('fr')
        fr_match = fr.acceptability(accept)
        self.assertEqual(fr_match.quality, 1.0)
        self.assertTrue(fr_match.exact_match)

        fr_be = Language('fr-be')
        fr_be_match = fr_be.acceptability(accept)
        self.assertEqual(fr_be_match.quality, 1.0)
        self.assertTrue(fr_be_match.exact_match)

        # more specific first
        self.assertGreater(fr_be_match, fr_match)

        en_gb = Language('en-gb')
        en_gb_match = en_gb.acceptability(accept)
        self.assertEqual(en_gb_match.quality, 0.8)
        self.assertTrue(en_gb_match.exact_match)

        en_us = Language('en-us')
        en_us_match = en_us.acceptability(accept)
        self.assertEqual(en_us_match.quality, 0.7)
        self.assertFalse(en_us_match.exact_match)

        en = Language('en')
        en_match = en.acceptability(accept)
        self.assertEqual(en_match.quality, 0.7)
        self.assertTrue(en_match.exact_match)

        zu = Language('zu')
        zu_match = zu.acceptability(accept)
        self.assertEqual(zu_match.quality, 0.1)
        self.assertFalse(zu_match.exact_match)



"""
    verktyg.testsuite.utils
    ~~~~~~~~~~~~~~~~~~~~~~~

    General utilities.

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

import inspect

from verktyg.responses import BaseResponse
from verktyg.test import Client
from verktyg import utils


class UtilsTestCase(unittest.TestCase):
    def test_redirect(self):
        resp = utils.redirect(u'/füübär')
        self.assertIn(b'/f%C3%BC%C3%BCb%C3%A4r', resp.get_data())
        self.assertEqual(resp.headers['Location'], '/f%C3%BC%C3%BCb%C3%A4r')
        self.assertEqual(resp.status_code, 302)

        resp = utils.redirect(u'http://☃.net/', 307)
        self.assertIn(b'http://xn--n3h.net/', resp.get_data())
        self.assertEqual(resp.headers['Location'], 'http://xn--n3h.net/')
        self.assertEqual(resp.status_code, 307)

        resp = utils.redirect('http://example.com/', 305)
        self.assertEqual(resp.headers['Location'], 'http://example.com/')
        self.assertEqual(resp.status_code, 305)

    def test_redirect_no_unicode_header_keys(self):
        # Make sure all headers are native keys.  This was a bug at one point
        # due to an incorrect conversion.
        resp = utils.redirect('http://example.com/', 305)
        for key, value in resp.headers.items():
            self.assertEqual(type(key), str)
            self.assertEqual(type(value), str)
        self.assertEqual(resp.headers['Location'], 'http://example.com/')
        self.assertEqual(resp.status_code, 305)

    def test_redirect_xss(self):
        location = 'http://example.com/?xss="><script>alert(1)</script>'
        resp = utils.redirect(location)
        self.assertNotIn(b'<script>alert(1)</script>', resp.get_data())

        location = 'http://example.com/?xss="onmouseover="alert(1)'
        resp = utils.redirect(location)
        self.assertNotIn(
            b'href="http://example.com/?xss="onmouseover="alert(1)"',
            resp.get_data()
        )

    def test_redirect_with_custom_response_class(self):
        class MyResponse(BaseResponse):
            pass

        location = "http://example.com/redirect"
        resp = utils.redirect(location, Response=MyResponse)

        self.assertIsInstance(resp, MyResponse)
        self.assertEqual(resp.headers['Location'], location)

    def test_cached_property(self):
        foo = []

        class A(object):

            def prop(self):
                foo.append(42)
                return 42
            prop = utils.cached_property(prop)

        a = A()
        p = a.prop
        q = a.prop
        self.assertEqual(p, q, 42)
        self.assertEqual(foo, [42])

        foo = []

        class A(object):

            def _prop(self):
                foo.append(42)
                return 42
            prop = utils.cached_property(_prop, name='prop')
            del _prop

        a = A()
        p = a.prop
        q = a.prop
        self.assertEqual(p, q, 42)
        self.assertEqual(foo, [42])

    def test_can_set_cached_property(self):
        class A(object):

            @utils.cached_property
            def _prop(self):
                return 'cached_property return value'

        a = A()
        a._prop = 'value'
        self.assertEqual(a._prop, 'value')

    def test_inspect_treats_cached_property_as_property(self):
        class A(object):

            @utils.cached_property
            def _prop(self):
                return 'cached_property return value'

        attrs = inspect.classify_class_attrs(A)
        for attr in attrs:
            if attr.name == '_prop':
                break
        self.assertEqual(attr.kind, 'property')

    def test_append_slash_redirect(self):
        def app(env, sr):
            return utils.append_slash_redirect(env)(env, sr)
        client = Client(app, BaseResponse)
        response = client.get('foo', base_url='http://example.org/app')
        self.assertEqual(
            response.status_code, 301
        )
        self.assertEqual(
            response.headers['Location'], 'http://example.org/app/foo/'
        )

    def test_cached_property_doc(self):
        @utils.cached_property
        def foo():
            """testing"""
            return 42
        self.assertEqual(foo.__doc__, 'testing')
        self.assertEqual(foo.__name__, 'foo')
        self.assertEqual(foo.__module__, __name__)

    def test_secure_filename(self):
        self.assertEqual(
            utils.secure_filename('My cool movie.mov'),
            'My_cool_movie.mov'
        )
        self.assertEqual(
            utils.secure_filename('../../../etc/passwd'),
            'etc_passwd'
        )
        self.assertEqual(
            utils.secure_filename(u'i contain cool \xfcml\xe4uts.txt'),
            'i_contain_cool_umlauts.txt'
        )



"""
    verktyg.testsuite.views
    ~~~~~~~~~~~~~~~~~~~~~~~

    Tests for view wrappers.

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import unittest

import verktyg
from verktyg.exceptions import MethodNotAllowed
import verktyg.views as vtv


class ViewsTestCase(unittest.TestCase):
    def test_decorators(self):
        dispatcher = verktyg.Dispatcher()

        @vtv.expose(dispatcher, 'foo')
        def foo(env, req):
            pass

    def test_class_view(self):
        dispatcher = verktyg.Dispatcher()

        class Foo(vtv.ClassView):
            name = 'foo'

            def GET(self, env, req):
                return 'get'

            def POST(self, env, req):
                return 'post'

        dispatcher.add_bindings(Foo())

        self.assertEqual(
            'get',
            dispatcher.lookup('foo', method='GET')(None, None)
        )
        self.assertEqual(
            'post',
            dispatcher.lookup('foo', method='POST')(None, None)
        )
        self.assertRaises(
            MethodNotAllowed,
            dispatcher.lookup, 'foo', method='PUT'
        )



"""
    verktyg.accept.language
    ~~~~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2016 Ben Mather
    :license:
        BSD, see LICENSE for more details.
"""
from verktyg.exceptions import NotAcceptable

from verktyg.accept import _base


class _LanguageRange(_base.Range):
    def _validate_param(self, key, value):
        raise ValueError("Accept-Language header does not take parameters")


class LanguageAccept(_base.Accept):
    range_type = _LanguageRange


class LanguageAcceptability(_base.Acceptability):
    def __init__(
        self, content_type, *,
        specificity, tail,
        q, qs=None
    ):
        super(LanguageAcceptability, self).__init__(
            content_type, match_quality=(-tail, specificity), q=q, qs=qs
        )

    @property
    def language(self):
        return self._value

    @property
    def specificity(self):
        return self._match_quality[1]

    @property
    def tail(self):
        return -self._match_quality[0]

    @property
    def exact_match(self):
        return not self.tail


class Language(_base.Value):
    match_type = LanguageAcceptability

    def _acceptability_for_option(self, option):
        if self.value == option.value:
            specificity = len(list(option.value.split('-')))
            tail = 0
        elif self.value.startswith('%s-' % option.value):
            specificity = len(list(option.value.split('-')))
            tail = len(list(self.value.split('-'))) - specificity
        elif option.value == '*':
            specificity = 0
            tail = len(list(self.value.split('-')))
        else:
            raise NotAcceptable()

        return self.match_type(
            self, specificity=specificity, tail=tail,
            qs=self.qs, q=option.q
        )


def parse_accept_language_header(string):
    return LanguageAccept(_base.split_accept_string(string))


def parse_language_header(string):
    return Language(string)



"""
    verktyg.accept.charset
    ~~~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2016 Ben Mather
    :license:
        BSD, see LICENSE for more details.
"""
from verktyg.accept import _base


class _CharsetRange(_base.Range):
    def _validate_param(self, key, value):
        raise ValueError("Accept-Charset header does not take parameters")


class CharsetAccept(_base.Accept):
    range_type = _CharsetRange


class CharsetAcceptability(_base.Acceptability):
    def __init__(self, value, *, exact_match, q, qs=None):
        super(CharsetAcceptability, self).__init__(
            value, match_quality=exact_match, q=q, qs=qs
        )

    @property
    def charset(self):
        return self._value

    @property
    def exact_match(self):
        return self._match_quality


class Charset(_base.Value):
    match_type = CharsetAcceptability


def parse_accept_charset_header(string):
    return CharsetAccept(_base.split_accept_string(string))


def parse_charset_header(string):
    return Charset(string)
